import torch
import torch.nn.functional as F
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


# reference https://github.com/tim-learn/ATDOC/blob/main/demo_uda.py
class NeighborhoodAggregation(torch.nn.Module):
    def __init__(self, dataset_size, feature_dim, num_classes, k=5, T=0.5):
        super().__init__()
        self.register_buffer(
            "feat_memory", F.normalize(torch.rand(dataset_size, feature_dim))
        )
        self.register_buffer(
            "pred_memory", torch.ones(dataset_size, num_classes) / num_classes
        )
        self.k = k
        self.T = T

    def forward(self, features, logits=None, update=False, idx=None):
        # move to device if necessary
        self.feat_memory = pml_cf.to_device(self.feat_memory, features)
        self.pred_memory = pml_cf.to_device(self.pred_memory, features)
        with torch.no_grad():
            features = F.normalize(features)
            pseudo_labels, mean_logits = self.get_pseudo_labels(features, idx)
            if update:
                self.update_memory(features, logits, idx)
        return pseudo_labels, mean_logits

    def get_pseudo_labels(self, normalized_features, idx):
        dis = torch.mm(normalized_features, self.feat_memory.t())
        # set self-comparisons to min similarity
        for di in range(dis.size(0)):
            dis[di, idx[di]] = torch.min(dis)
        _, indices = torch.topk(dis, k=self.k, dim=1)
        logits = torch.mean(self.pred_memory[indices], dim=1)
        pseudo_labels = torch.argmax(logits, dim=1)
        return pseudo_labels, logits

    def update_memory(self, normalized_features, logits, idx):
        logits = F.softmax(logits, dim=1)
        p = 1.0 / self.T
        logits = (logits ** p) / torch.sum(logits ** p, dim=1, keepdims=True)
        self.feat_memory[idx] = normalized_features
        self.pred_memory[idx] = logits

    def extra_repr(self):
        return c_f.extra_repr(self, ["k", "T"])



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


def entropy(logits):
    entropies = -torch.sum(
        torch.softmax(logits, dim=1) * torch.log_softmax(logits, dim=1), dim=1
    )
    return entropies


def entropy_after_softmax(preds, eps=None):
    eps = c_f.default(eps, pml_cf.small_val(preds.dtype))
    entropies = -torch.sum(preds * torch.log(preds + eps), dim=1)
    return entropies


def get_entropy(logits, after_softmax):
    if after_softmax:
        return entropy_after_softmax(logits)
    return entropy(logits)


class EntropyLoss(torch.nn.Module):
    def __init__(self, after_softmax=False, return_mean=True):
        super().__init__()
        self.after_softmax = after_softmax
        self.return_mean = return_mean

    def forward(self, logits):
        entropies = get_entropy(logits, self.after_softmax)
        if self.return_mean:
            return torch.mean(entropies)
        return entropies

    def extra_repr(self):
        return c_f.extra_repr(self, ["after_softmax"])



import copy
from collections import defaultdict

from ..utils import common_functions as c_f
from .base import BaseHook, check_keys_are_present, replace_mapped_keys
from .features import BaseFeaturesHook
from .utils import ChainHook, ParallelHook, RepeatHook


def update_model_counts(hook, available_keys, model_counts):
    # get model counts
    # will be correct only for non-conditional hooks
    if isinstance(hook, BaseFeaturesHook):
        needs_compute = set(hook.out_keys) - available_keys
        needs_compute_by_domain = defaultdict(list)
        for nc in needs_compute:
            needs_compute_by_domain[nc.split("_")[0]].append(nc)
        for domain, ncs in needs_compute_by_domain.items():
            if not hook.detach[domain]:
                model_counts[hook.model_name] += 1
            else:
                for nc in ncs:
                    detachable_regex = hook.detachable_regex[nc]
                    if not any(detachable_regex.search(ak) for ak in available_keys):
                        model_counts[hook.model_name] += 1
                        break


def validate_hook(hook, available_keys=None, depth=0, model_counts=None):
    c_f.LOGGER.debug(f"VALIDATE: {'  '*depth}{c_f.cls_name(hook)}")
    available_keys = c_f.default(available_keys, [])
    model_counts = c_f.default(model_counts, defaultdict(int))

    if isinstance(available_keys, list):
        available_keys = set(available_keys)

    if isinstance(hook, ChainHook):
        hooks = hook.hooks
        for i in range(0, len(hooks)):
            validate_hook(hooks[i], available_keys, depth + 1, model_counts)

    elif isinstance(hook, ParallelHook):
        hooks = hook.hooks
        for i in range(0, len(hooks)):
            curr_available_keys = copy.deepcopy(available_keys)
            validate_hook(hooks[i], curr_available_keys, depth + 1, model_counts)

    elif isinstance(hook, RepeatHook):
        for _ in range(hook.n):
            curr_available_keys = copy.deepcopy(available_keys)
            validate_hook(hook.hook, curr_available_keys, depth + 1, model_counts)

    else:
        check_keys_are_present(
            hook, hook.in_keys, list(available_keys), "in_keys", "available_keys"
        )
        check_keys_are_present(
            hook,
            list(hook.key_map.keys()),
            list(available_keys),
            "key_map",
            "available_keys",
        )
        all_hooks = c_f.attrs_of_type(hook, BaseHook)
        for h in all_hooks.values():
            validate_hook(h, available_keys, depth + 1, model_counts)
        update_model_counts(hook, available_keys, model_counts)
        available_keys.update(set(hook.out_keys))

    return model_counts



import torch

from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .features import (
    FeaturesAndLogitsHook,
    FeaturesChainHook,
    FeaturesHook,
    FrozenModelHook,
    LogitsHook,
)
from .optimizer import OptimizerHook, SummaryHook
from .utils import ApplyFnHook, ChainHook, OnlyNewOutputsHook


class SoftmaxHook(ApplyFnHook):
    def __init__(self, **kwargs):
        super().__init__(fn=torch.nn.Softmax(dim=1), **kwargs)


class SoftmaxLocallyHook(BaseWrapperHook):
    def __init__(self, apply_to, *hooks, **kwargs):
        super().__init__(**kwargs)
        s_hook = SoftmaxHook(apply_to=apply_to)
        self.hook = OnlyNewOutputsHook(ChainHook(s_hook, *hooks, overwrite=True))


class CLossHook(BaseWrapperHook):
    def __init__(self, loss_fn=None, detach_features=False, f_hook=None, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = c_f.default(
            loss_fn, torch.nn.CrossEntropyLoss, {"reduction": "none"}
        )
        self.hook = c_f.default(
            f_hook,
            FeaturesAndLogitsHook,
            {"domains": ["src"], "detach_features": detach_features},
        )

    def call(self, losses, inputs):
        outputs = self.hook(losses, inputs)[1]
        [src_logits] = c_f.extract(
            [outputs, inputs], c_f.filter(self.hook.out_keys, "_logits$")
        )
        loss = self.loss_fn(src_logits, inputs["src_labels"])
        return {self._loss_keys()[0]: loss}, outputs

    def _loss_keys(self):
        return ["c_loss"]


class ClassifierHook(BaseWrapperHook):
    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        loss_fn=None,
        f_hook=None,
        detach_features=False,
        pre=None,
        post=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        [pre, post] = c_f.many_default([pre, post], [[], []])
        hook = CLossHook(loss_fn, detach_features, f_hook)
        hook = ChainHook(*pre, hook, *post)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)


class FinetunerHook(ClassifierHook):
    def __init__(self, **kwargs):
        f_hook = FrozenModelHook(FeaturesHook(detach=True, domains=["src"]), "G")
        f_hook = FeaturesChainHook(f_hook, LogitsHook(domains=["src"]))
        super().__init__(f_hook=f_hook, **kwargs)



import torch


class PlusResidual(torch.nn.Module):
    def __init__(self, residual):
        super().__init__()
        self.residual = residual

    def forward(self, x):
        return x + self.residual(x)



from ..layers import GradientReversal
from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .classification import CLossHook, SoftmaxHook
from .domain import DomainLossHook, FeaturesForDomainLossHook
from .features import FeaturesHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import (
    ApplyFnHook,
    AssertHook,
    BaseWrapperHook,
    ChainHook,
    OnlyNewOutputsHook,
)


class GradientReversalHook(ApplyFnHook):
    def __init__(self, **kwargs):
        super().__init__(fn=GradientReversal(), **kwargs)


class SoftmaxGradientReversalHook(BaseWrapperHook):
    def __init__(self, apply_to, **kwargs):
        super().__init__(**kwargs)
        self.hook = ChainHook(
            SoftmaxHook(apply_to=apply_to),
            GradientReversalHook(apply_to=apply_to),
            overwrite=True,
        )


class DANNHook(BaseWrapperHook):
    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        pre=None,
        pre_d=None,
        post_d=None,
        pre_g=None,
        post_g=None,
        gradient_reversal=None,
        use_logits=False,
        f_hook=None,
        d_hook=None,
        c_hook=None,
        domain_loss_hook=None,
        d_hook_allowed="_dlogits$",
        **kwargs
    ):
        super().__init__(**kwargs)
        [pre, pre_d, post_d, pre_g, post_g] = c_f.many_default(
            [pre, pre_d, post_d, pre_g, post_g], [[], [], [], [], []]
        )
        f_hook = c_f.default(
            f_hook, FeaturesForDomainLossHook, {"use_logits": use_logits}
        )
        gradient_reversal = c_f.default(
            gradient_reversal, GradientReversalHook, {"apply_to": f_hook.out_keys}
        )
        c_hook = c_f.default(c_hook, CLossHook, {})
        domain_loss_hook = c_f.default(
            domain_loss_hook, DomainLossHook, {"f_hook": f_hook, "d_hook": d_hook}
        )

        disc_hook = AssertHook(
            OnlyNewOutputsHook(
                ChainHook(
                    gradient_reversal,
                    *pre_d,
                    domain_loss_hook,
                    *post_d,
                    overwrite=[1],
                )
            ),
            d_hook_allowed,
        )
        gen_hook = ChainHook(*pre_g, c_hook, *post_g)

        hook = ChainHook(*pre, f_hook, disc_hook, gen_hook)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)


class DANNLogitsHook(DANNHook):
    def __init__(self, **kwargs):
        f_hook = FeaturesForDomainLossHook(use_logits=True)
        super().__init__(f_hook=f_hook, **kwargs)


class DANNSoftmaxLogitsHook(DANNHook):
    def __init__(self, **kwargs):
        f_hook = FeaturesForDomainLossHook(use_logits=True)
        gradient_reversal = SoftmaxGradientReversalHook(apply_to=f_hook.out_keys)
        super().__init__(f_hook=f_hook, gradient_reversal=gradient_reversal, **kwargs)



import torch
import torch.nn.functional as F

from ..utils import common_functions as c_f


def mcd_loss(out1, out2, dist_fn):
    return dist_fn(F.softmax(out1, dim=1), F.softmax(out2, dim=1))


class MCDLoss(torch.nn.Module):
    def __init__(self, dist_fn=None):
        super().__init__()
        self.dist_fn = c_f.default(dist_fn, torch.nn.L1Loss, {})

    def forward(self, x, y):
        return mcd_loss(x, y, self.dist_fn)


def general_mcd_loss(*x, p=1):
    x = [torch.nn.functional.softmax(i, dim=1) for i in x]
    loss = 0
    for i in range(len(x)):
        for j in range(i + 1, len(x)):
            loss += F(x[i], x[j])
    return loss


class GeneralMCDLoss(torch.nn.Module):
    def __init__(self, p=1):
        super().__init__()
        self.p = p

    def forward(self, *x):
        return general_mcd_loss(*x, p=self.p)



import torch

from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..hooks import ClassifierHook, FinetunerHook
from .base_adapter import BaseGCAdapter
from .utils import default_optimizer_tuple


class Classifier(BaseGCAdapter):
    hook_cls = ClassifierHook

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(opts=list(self.optimizers.values()), **hook_kwargs)


class Finetuner(Classifier):
    hook_cls = FinetunerHook

    def get_default_containers(self):
        optimizers = Optimizers(default_optimizer_tuple(), keys=["C"])
        return MultipleContainers(optimizers=optimizers)

    def get_key_enforcer(self):
        ke = super().get_key_enforcer()
        ke.requirements["optimizers"].remove("G")
        return ke



import math

import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


# modified from https://github.com/thuml/CDAN/blob/master/pytorch/network.py
class RandomizedDotProduct(torch.nn.Module):
    def __init__(self, in_dims, out_dim=1024):
        super().__init__()
        self.in_dims = in_dims
        for i, d in enumerate(in_dims):
            self.register_buffer(self.rand_mat_name(i), torch.randn(d, out_dim))
        self.out_dim = out_dim
        self.num_mats = len(in_dims)
        self.divisor = math.pow(float(self.out_dim), 1.0 / self.num_mats)

    def forward(self, *inputs):
        for i in range(self.num_mats):
            # move to device if necessary
            curr = inputs[i]
            self.set_rand_mat(
                i, pml_cf.to_device(self.get_rand_mat(i), curr, dtype=curr.dtype)
            )

        return_list = [
            torch.mm(inputs[i], self.get_rand_mat(i)) for i in range(self.num_mats)
        ]
        return_tensor = return_list[0] / self.divisor
        for single in return_list[1:]:
            return_tensor = return_tensor * single
        return return_tensor

    def set_rand_mat(self, i, value):
        setattr(self, self.rand_mat_name(i), value)

    def get_rand_mat(self, i):
        return getattr(self, self.rand_mat_name(i))

    def rand_mat_name(self, i):
        return f"rand_mat{i}"

    def extra_repr(self):
        return c_f.extra_repr(self, ["in_dims", "out_dim", "divisor"])



import torch


class DoNothingOptimizer(torch.optim.Optimizer):
    def __init__(self, *arg, **kwargs):
        self.param_groups = [{"lr": 0}]

    def step(self):
        pass

    def zero_grad(self):
        pass

    def state_dict(self):
        return {}

    def load_state_dict(self, state_dict):
        pass

    def __repr__(self):
        return "DoNothingOptimizer()"



from .adabn import AdaBN
from .adda import ADDA
from .aligner import RTN, Aligner
from .base_adapter import BaseAdapter
from .classifier import Classifier, Finetuner
from .dann import DANN, GVB
from .gan import CDAN, GAN, VADA, DomainConfusion
from .mcd import MCD
from .symnets import SymNets



from ..utils import common_functions as c_f
from ..weighters import MeanWeighter
from .base import BaseHook, BaseWrapperHook
from .reducers import MeanReducer


class OptimizerHook(BaseHook):
    # optimizers is a list of optimizers
    def __init__(self, hook, optimizers, weighter=None, reducer=None, **kwargs):
        super().__init__(**kwargs)
        self.hook = hook
        self.optimizers = optimizers
        self.weighter = c_f.default(weighter, MeanWeighter, {})
        self.reducer = c_f.default(reducer, MeanReducer, {})
        self.loss_components = {}

    def call(self, losses, inputs):
        losses, outputs = self.hook(losses, inputs)
        losses, new_outputs = self.reducer(losses, {**inputs, **outputs})
        outputs.update(new_outputs)
        loss, self.loss_components = self.weighter(losses)
        c_f.zero_back_step(loss, self.optimizers)
        return {}, outputs

    def _loss_keys(self):
        return []

    def _out_keys(self):
        return c_f.join_lists([self.hook.out_keys, self.reducer.out_keys])

    def extra_repr(self):
        return c_f.extra_repr(self, ["weighter"])


class SummaryHook(BaseHook):
    # optimizers is a dict of optimizer hooks
    def __init__(self, optimizers, **kwargs):
        super().__init__(**kwargs)
        self.optimizers = optimizers

    def call(self, losses, inputs):
        losses = {}
        for k, v in self.optimizers.items():
            losses[k] = v.loss_components
        return losses, {}

    def _loss_keys(self):
        return list(self.optimizers.keys())

    def _out_keys(self):
        return []



import copy

import torch

from ..utils import common_functions as c_f


class ModelWithBridge(torch.nn.Module):
    def __init__(self, model, bridge=None):
        super().__init__()
        self.model = model
        if bridge is None:
            bridge = c_f.reinit(copy.deepcopy(model))
        self.bridge = bridge

    def forward(self, x, return_bridge=False):
        y = self.model(x)
        z = self.bridge(x)
        output = y - z
        if return_bridge:
            return output, z
        return output



import torch

from ..utils import common_functions as c_f
from .adaptive_batch_norm import (
    AdaptiveBatchNorm2d,
    PopulationBatchNorm2d,
    convert_bn_to_adabn,
    finalize_bn,
    set_curr_domain,
)


class AdaBNModel(torch.nn.Module):
    def __init__(self, model, affine_domain=0, bn_type=None):
        super().__init__()
        bn_type = c_f.default(bn_type, torch.nn.BatchNorm2d)
        convert_bn_to_adabn(model, affine_domain=affine_domain, bn_type=bn_type)
        self.model = model

    def forward(self, x, domain):
        domain = c_f.check_domain(self, domain)
        set_curr_domain(self.model, domain, AdaptiveBatchNorm2d)
        return self.model(x)

    def eval(self):
        super().eval()
        finalize_bn(self, PopulationBatchNorm2d)



from ..containers import KeyEnforcer, MultipleContainers
from ..hooks import AdaBNHook
from ..utils.common_functions import check_domain
from .base_adapter import BaseAdapter


class AdaBN(BaseAdapter):
    hook_cls = AdaBNHook

    def inference_default(self, x, domain):
        domain = check_domain(self, domain, keep_len=True)
        features = self.models["G"](x, domain)
        logits = self.models["C"](features, domain)
        return features, logits

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(**hook_kwargs)

    def get_key_enforcer(self):
        return KeyEnforcer(models=["G", "C"], optimizers=[])

    def get_default_containers(self):
        return MultipleContainers()



from .accuracy_validator import AccuracyValidator
from .base_validator import BaseValidator
from .deep_embedded_validator import DeepEmbeddedValidator
from .diversity_validator import DiversityValidator
from .entropy_validator import EntropyValidator
from .error_validator import ErrorValidator
from .knn_validator import ClusterValidator, KNNValidator
from .multiple_validators import MultipleValidators
from .silhouette_score_validator import SilhouetteScoreValidator
from .snd_validator import SNDValidator



import torch

from .domain_dataset import DomainDataset


class SourceDataset(DomainDataset):
    def __init__(self, dataset, domain=0):
        super().__init__(dataset, domain)

    def __getitem__(self, idx):
        img, src_labels = self.dataset[idx]
        return {
            "src_imgs": img,
            "src_domain": self.domain,
            "src_labels": src_labels,
            "sample_idx": idx,
        }



from .base_container import BaseContainer


class Models(BaseContainer):
    def train(self):
        for v in self.values():
            v.train()

    def eval(self):
        for v in self.values():
            v.eval()

    def zero_grad(self):
        for v in self.values():
            v.zero_grad()



import torch

from ..layers import DiversityLoss
from .base_validator import BaseValidator


class DiversityValidator(BaseValidator):
    def compute_score(self, target_train):
        return DiversityLoss()(target_train["logits"]).item()

    @property
    def maximize(self):
        return False



import torch

from ..utils import common_functions as c_f
from .normalizers import NoNormalizer


class ConfidenceWeights(torch.nn.Module):
    def __init__(self, normalizer=None):
        super().__init__()
        self.normalizer = c_f.default(normalizer, NoNormalizer())

    def forward(self, logits):
        return self.normalizer(torch.max(logits, dim=1)[0])



import torch

from ..utils import common_functions as c_f


class ConcatSoftmax(torch.nn.Module):
    def __init__(self, dim=1):
        super().__init__()
        self.dim = dim

    def forward(self, *x):
        all_logits = torch.cat(x, dim=self.dim)
        return torch.nn.functional.softmax(all_logits, dim=self.dim)

    def extra_repr(self):
        return c_f.extra_repr(self, ["dim"])



import torch

from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .classification import SoftmaxLocallyHook
from .domain import DomainLossHook
from .features import CombinedFeaturesHook, FeaturesAndLogitsHook
from .gan import GANHook
from .utils import ChainHook


def cdan_key_map(fc_hook):
    strs = c_f.filter(fc_hook.out_keys, "", ["^src", "^target"])
    return {k: v for k, v in zip(strs, ["src_imgs_features", "target_imgs_features"])}


class CDANDomainHook(BaseWrapperHook):
    def __init__(
        self, loss_prefix, detach_features, reverse_labels, softmax=True, **kwargs
    ):
        super().__init__(**kwargs)
        f_hook = FeaturesAndLogitsHook()
        fc_hook = CombinedFeaturesHook()
        key_map = cdan_key_map(fc_hook)
        if softmax:
            strs = c_f.filter(f_hook.out_keys, "_features_logits$", ["^src", "^target"])
            fc_hook = SoftmaxLocallyHook(strs, fc_hook)
        d_hook = DomainLossHook(
            loss_prefix=loss_prefix,
            detach_features=detach_features,
            reverse_labels=reverse_labels,
            key_map=key_map,
        )
        self.hook = ChainHook(f_hook, fc_hook, d_hook)


class CDANDomainHookD(CDANDomainHook):
    def __init__(self, **kwargs):
        super().__init__(
            loss_prefix="d_", detach_features=True, reverse_labels=False, **kwargs
        )


class CDANDomainHookG(CDANDomainHook):
    def __init__(self, **kwargs):
        super().__init__(
            loss_prefix="g_", detach_features=False, reverse_labels=True, **kwargs
        )


class CDANHook(GANHook):
    def __init__(self, softmax=True, **kwargs):
        super().__init__(
            disc_hook=CDANDomainHookD(softmax=softmax),
            gen_hook=CDANDomainHookG(softmax=softmax),
            **kwargs
        )



from abc import ABC, abstractmethod
from enum import Enum

import torch

from ..containers import KeyEnforcer, MultipleContainers, Optimizers
from ..utils import common_functions as c_f
from .utils import default_optimizer_tuple


class BaseAdapter(ABC):
    def __init__(
        self,
        models=None,
        optimizers=None,
        lr_schedulers=None,
        misc=None,
        default_containers=None,
        key_enforcer=None,
        inference=None,
        before_training_starts=None,
        hook_kwargs=None,
    ):
        self.containers = c_f.default(
            default_containers, self.get_default_containers, {}
        )
        self.key_enforcer = c_f.default(key_enforcer, self.get_key_enforcer, {})
        self.before_training_starts = c_f.class_default(
            self, before_training_starts, self.before_training_starts_default
        )

        self.containers.merge(
            models=models,
            optimizers=optimizers,
            lr_schedulers=lr_schedulers,
            misc=misc,
        )

        hook_kwargs = c_f.default(hook_kwargs, {})
        self.init_containers_and_check_keys()
        self.init_hook(hook_kwargs)
        self.inference = c_f.class_default(self, inference, self.inference_default)

    def training_step(self, batch, device, framework):
        batch = c_f.batch_to_device(batch, device)
        losses, _ = self.hook({}, {**self.models, **self.misc, **batch})
        return losses

    def inference_default(self, x, domain=None):
        features = self.models["G"](x)
        logits = self.models["C"](features)
        return features, logits

    def get_default_containers(self):
        optimizers = Optimizers(default_optimizer_tuple())
        return MultipleContainers(optimizers=optimizers)

    @abstractmethod
    def get_key_enforcer(self):
        pass

    @abstractmethod
    def init_hook(self):
        pass

    def init_containers_and_check_keys(self):
        self.containers.create()
        self.key_enforcer.check(self.containers)
        for k, v in self.containers.items():
            setattr(self, k, v)

    def before_training_starts_default(self, framework):
        c_f.LOGGER.info(f"models\n{self.models}")
        c_f.LOGGER.info(f"optimizers\n{self.optimizers}")
        c_f.LOGGER.info(f"lr_schedulers\n{self.lr_schedulers}")
        c_f.LOGGER.info(f"misc\n{self.misc}")
        c_f.LOGGER.info(f"hook\n{self.hook}")


class BaseGCDAdapter(BaseAdapter):
    def get_key_enforcer(self):
        return KeyEnforcer(
            models=["G", "C", "D"],
            optimizers=["G", "C", "D"],
        )


class BaseGCAdapter(BaseAdapter):
    def get_key_enforcer(self):
        return KeyEnforcer(
            models=["G", "C"],
            optimizers=["G", "C"],
        )



import torch.nn as nn


class Discriminator(nn.Module):
    def __init__(self, in_size=2048, h=2048, out_size=1):
        super().__init__()
        self.h = h
        self.net = nn.Sequential(
            nn.Linear(in_size, h),
            nn.ReLU(),
            nn.Linear(h, h // 2),
            nn.ReLU(),
            nn.Linear(h // 2, out_size),
        )
        self.out_size = out_size

    def forward(self, x):
        return self.net(x).squeeze(1)



import copy

import torch


class MultipleModels(torch.nn.Module):
    def __init__(self, *models):
        super().__init__()
        self.models = torch.nn.ModuleList(models)

    def forward(self, x):
        outputs = []
        for m in self.models:
            outputs.append(m(x))
        return outputs



import math

import torch

from ..utils import common_functions as c_f


# adapted from https://github.com/jihanyang/AFN/blob/master/vanilla/Office31/SAFN/code/train.py
class AdaptiveFeatureNorm(torch.nn.Module):
    def __init__(self, step_size=1):
        super().__init__()
        self.step_size = step_size

    def forward(self, x):
        l2_norm = x.norm(p=2, dim=1)
        radius = l2_norm.detach() + self.step_size
        return torch.mean((l2_norm - radius) ** 2)

    def extra_repr(self):
        return c_f.extra_repr(self, ["step_size"])


class L2PreservedDropout(torch.nn.Module):
    def __init__(self, p=0.5, inplace=False):
        super().__init__()
        self.dropout = torch.nn.Dropout(p=p, inplace=inplace)
        self.scale = math.sqrt(1 - p)

    def forward(self, x):
        x = self.dropout(x)
        if self.training:
            return x * self.scale
        return x



from collections.abc import MutableMapping

from ..utils import common_functions as c_f


# https://stackoverflow.com/a/3387975
class BaseContainer(MutableMapping):
    def __init__(self, store, other=None, keys=None):
        if not isinstance(store, (tuple, dict)):
            raise TypeError("BaseContainer input must be a tuple or dict")
        if isinstance(store, tuple):
            self.store_as_tuple = store
            self.store = {}
        else:
            self.store_as_tuple = None
            self.store = store
        if keys is not None:
            self.duplicate(keys)
        if other is not None:
            self.create_with(other)

    def __getitem__(self, key):
        return self.store[self._keytransform(key)]

    def __setitem__(self, key, value):
        self.store[self._keytransform(key)] = value

    def __delitem__(self, key):
        del self.store[self._keytransform(key)]

    def __iter__(self):
        return iter(self.store)

    def __len__(self):
        return len(self.store)

    def _keytransform(self, key):
        return key

    def __repr__(self):
        if isinstance(self.store, dict):
            output = ""
            for k, v in self.items():
                output += f"{k}: {v}\n"
            return output
        return str(self.store)

    def merge(self, other):
        if not isinstance(other, BaseContainer):
            raise TypeError("merge can only be done with another container")
        if other.store_as_tuple:
            if len(self) > 0:
                for k, v in self.items():
                    self[k] = other.store_as_tuple
            else:
                self.store_as_tuple = other.store_as_tuple
        else:
            for k, v in other.items():
                self[k] = v

    def create(self):
        for k, v in self.items():
            if isinstance(v, tuple):
                if len(v) == 2:
                    class_ref, kwargs = v
                    self[k] = class_ref(**kwargs)
                elif len(v) == 1:
                    self[k] = v[0]
                else:
                    raise ValueError(
                        f"The tuple {v} has length={len(v)}, but it must be of length 1 or 2"
                    )
        self.delete_unwanted_keys()

    def create_with(self, other):
        self.store_as_tuple = self.type_check(self.store_as_tuple, other)
        self.store = self.type_check(self.store, other)
        self.store_as_tuple.update(self.store)
        self.store = self.store_as_tuple
        self.store_as_tuple = None
        self.delete_unwanted_keys()
        self._create_with(other)

    def _create_with(self, other):
        pass

    def type_check(self, store, other):
        if isinstance(store, tuple):
            return {k: store for k in other.keys()}
        elif isinstance(store, dict):
            return store
        elif store is None:
            return {}

    def duplicate(self, keys):
        if isinstance(self.store_as_tuple, tuple):
            self.store = {k: self.store_as_tuple for k in keys}
            self.store_as_tuple = None
        else:
            raise TypeError("If keys are specified, store must be a tuple.")

    def apply(self, function, keys=None):
        if keys is None:
            keys = list(self.keys())
        for k in keys:
            self[k] = function(self[k])

    def delete_unwanted_keys(self):
        del_list = []
        for k, v in self.items():
            if isinstance(v, DeleteKey):
                del_list.append(k)
        for k in del_list:
            del self[k]


class KeyEnforcer:
    def __init__(self, **kwargs):
        self.requirements = kwargs

    def check(self, containers):
        for k, required_keys in self.requirements.items():
            container_keys = list(containers[k].keys())
            r_c_diff = c_f.list_diff(required_keys, container_keys)
            c_r_diff = c_f.list_diff(container_keys, required_keys)
            error_msg = ""
            if len(r_c_diff) > 0:
                error_msg += (
                    f"The {k} container is missing the following keys: {r_c_diff}. "
                )

            if len(c_r_diff) > 0:
                error_msg += (
                    f"The {k} container has the following unallowed keys: {c_r_diff}."
                )

            if error_msg != "":
                raise KeyError(error_msg)


class DeleteKey:
    pass


def containers_are_equal(x, y):
    if not isinstance(x, BaseContainer) or not isinstance(y, BaseContainer):
        raise TypeError("inputs must be Container types")
    if x.keys() != y.keys():
        return False
    for k, v1 in x.items():
        v2 = y[k]
        if type(v1) != type(v2):
            return False
        elif isinstance(v1, BaseContainer):
            if not containers_are_equal(v1, v2):
                return False
        elif not c_f.has_state_dict(v1):
            if v1.__dict__ != v2.__dict__:
                return False
        elif c_f.has_state_dict(v1):
            if not c_f.state_dicts_are_equal(v1.state_dict(), v2.state_dict()):
                return False
    return True



import torch

from ..utils import common_functions as c_f
from .entropy_loss import get_entropy
from .normalizers import sum_normalizer


def entropy_weights(logits, after_softmax, normalizer):
    entropies = get_entropy(logits, after_softmax)
    weights = 1 + torch.exp(-entropies)
    return normalizer(weights)


class EntropyWeights(torch.nn.Module):
    def __init__(self, after_softmax=False, normalizer=None):
        super().__init__()
        self.after_softmax = after_softmax
        self.normalizer = c_f.default(normalizer, sum_normalizer)

    def forward(self, logits):
        return entropy_weights(logits, self.after_softmax, self.normalizer)

    def extra_repr(self):
        return c_f.extra_repr(self, ["after_softmax"])



from ..layers import MMDLoss
from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .classification import CLossHook, SoftmaxLocallyHook
from .features import FeaturesAndLogitsHook, FeaturesHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import ChainHook


class AlignerHook(BaseWrapperHook):
    def __init__(self, loss_fn=None, hook=None, layer="features", **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = c_f.default(loss_fn, MMDLoss, {})
        if layer == "features":
            default_hook = FeaturesHook
        elif layer == "logits":
            default_hook = FeaturesAndLogitsHook
        else:
            raise ValueError("AlignerHook layer must be 'features' or 'logits'")
        self.hook = c_f.default(hook, default_hook, {})
        self.layer = layer

    def call(self, losses, inputs):
        outputs = self.hook(losses, inputs)[1]
        strs = c_f.filter(self.hook.out_keys, f"_{self.layer}$", ["^src", "^target"])
        [src, target] = c_f.extract([outputs, inputs], strs)
        confusion_loss = self.loss_fn(src, target)
        return {self._loss_keys()[0]: confusion_loss}, outputs

    def _loss_keys(self):
        return [f"{self.layer}_confusion_loss"]


class JointAlignerHook(BaseWrapperHook):
    def __init__(self, loss_fn=None, hook=None, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = c_f.default(loss_fn, MMDLoss, {})
        self.hook = c_f.default(hook, FeaturesAndLogitsHook, {})

    def call(self, losses, inputs):
        outputs = self.hook(losses, inputs)[1]
        src = self.get_all_domain_features(inputs, outputs, "src")
        target = self.get_all_domain_features(inputs, outputs, "target")
        confusion_loss = self.loss_fn(src, target)
        return {self._loss_keys()[0]: confusion_loss}, outputs

    def _loss_keys(self):
        return ["joint_confusion_loss"]

    def get_all_domain_features(self, inputs, outputs, domain):
        return c_f.extract(
            [outputs, inputs], sorted(c_f.filter(self.hook.out_keys, f"^{domain}"))
        )


class FeaturesLogitsAlignerHook(BaseWrapperHook):
    def __init__(self, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        loss_fn = c_f.default(loss_fn, MMDLoss, {})
        features_hook = FeaturesAndLogitsHook()
        a1_hook = AlignerHook(loss_fn, layer="features")
        a2_hook = AlignerHook(loss_fn, layer="logits")
        self.hook = ChainHook(a1_hook, a2_hook)


class ManyAlignerHook(BaseWrapperHook):
    def __init__(
        self,
        loss_fn=None,
        aligner_hook=None,
        features_hook=None,
        softmax=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        loss_fn = c_f.default(loss_fn, MMDLoss, {})
        features_hook = c_f.default(features_hook, FeaturesAndLogitsHook, {})
        aligner_hook = c_f.default(
            aligner_hook, FeaturesLogitsAlignerHook, {"loss_fn": loss_fn}
        )
        if softmax:
            apply_to = c_f.filter(features_hook.out_keys, "_logits$")
            aligner_hook = SoftmaxLocallyHook(apply_to, aligner_hook)
        self.hook = ChainHook(features_hook, aligner_hook)


class AlignerPlusCHook(BaseWrapperHook):
    def __init__(
        self,
        opts,
        weighter=None,
        reducer=None,
        loss_fn=None,
        aligner_hook=None,
        pre=None,
        post=None,
        softmax=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        [pre, post] = c_f.many_default([pre, post], [[], []])
        aligner_hook = ManyAlignerHook(
            loss_fn=loss_fn, aligner_hook=aligner_hook, softmax=softmax
        )
        hook = ChainHook(*pre, aligner_hook, CLossHook(), *post)
        hook = OptimizerHook(hook, opts, weighter, reducer)
        s_hook = SummaryHook({"total_loss": hook})
        self.hook = ChainHook(hook, s_hook)



import torch

from .domain_dataset import DomainDataset


class PseudoLabeledDataset(DomainDataset):
    def __init__(self, dataset, pseudo_labels, domain=0):
        super().__init__(dataset, domain)
        if len(self.dataset) != len(pseudo_labels):
            raise ValueError("len(dataset) must equal len(pseudo_labels)")
        self.pseudo_labels = pseudo_labels

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, idx):
        img, _ = self.dataset[idx]
        return {
            "src_imgs": img,
            "src_domain": self.domain,
            "src_labels": self.pseudo_labels[idx],
            "sample_idx": idx,
        }



import os

import ignite.distributed as idist
import torch
from ignite.engine import Engine, Events
from ignite.handlers import TerminateOnNan

from ...datasets import DataloaderCreator
from ...utils import common_functions as c_f
from . import utils as i_g
from .loggers import IgniteEmptyLogger


class Ignite:
    def __init__(self, adapter, logger=None, log_freq=50, with_pbars=True):
        self.adapter = adapter
        self.logger = c_f.default(logger, IgniteEmptyLogger, {})
        self.log_freq = log_freq
        self.with_pbars = with_pbars
        self.engine_init()
        self.dist_init_done = False

    def training_step(self, engine, batch):
        device = idist.device()
        batch = c_f.batch_to_device(batch, device)
        return self.adapter.training_step(batch, device, self)

    def before_training_starts(self, engine):
        self.adapter.before_training_starts(self)

    def engine_init(self):
        self.trainer = Engine(self.training_step)
        self.trainer.state.adapter = self.adapter
        self.labeled_collector = Engine(
            self.get_labeled_collector_step(self.adapter.inference)
        )
        self.unlabeled_collector = Engine(
            self.get_unlabeled_collector_step(self.adapter.inference)
        )
        i_g.do_for_all_engines(self, i_g.set_engine_logger)
        self.register_event_handlers()

    def register_event_handlers(self):
        if self.with_pbars:
            pbars = i_g.do_for_all_engines(self, i_g.attach_pbar)
        i_g.register(self.trainer, Events.STARTED, self.before_training_starts)
        i_g.register(
            self.trainer, Events.EPOCH_STARTED, self.set_to_train(self.adapter.models)
        )
        iteration_complete = [
            i_g.step_lr_schedulers(self.adapter.lr_schedulers, "per_step"),
            TerminateOnNan(),
        ]
        if self.with_pbars:
            iteration_complete.append(i_g.pbar_print_losses(pbars["trainer"]))
        i_g.register(self.trainer, Events.ITERATION_COMPLETED, *iteration_complete)
        i_g.register(
            self.trainer,
            Events.ITERATION_COMPLETED(every=self.log_freq),
            self.logger.add_training,
        )
        i_g.register(
            self.trainer,
            Events.EPOCH_COMPLETED,
            i_g.step_lr_schedulers(self.adapter.lr_schedulers, "per_epoch"),
            self.logger.write,
            i_g.zero_grad(self.adapter),
        )
        i_g.register(
            self.labeled_collector,
            Events.EPOCH_STARTED,
            self.set_to_eval(self.adapter.models),
        )
        i_g.register(
            self.unlabeled_collector,
            Events.EPOCH_STARTED,
            self.set_to_eval(self.adapter.models),
        )

    def dist_init(self, *args, **kwargs):
        if not self.dist_init_done:
            self.adapter.models.apply(i_g.auto_model(*args, **kwargs))
            self.adapter.optimizers.apply(idist.auto_optim)
            self.dist_init_done = True

    def get_training_length(self):
        max_epochs = self.trainer.state.max_epochs
        max_iters = max_epochs * self.trainer.state.epoch_length
        return max_epochs, max_iters

    def get_progress(self):
        _, max_iters = self.get_training_length()
        return float(self.get_iteration()) / max_iters

    def get_iteration(self):
        return self.trainer.state.iteration

    def get_epoch_length(self):
        return self.trainer.state.epoch_length

    def get_all_outputs(self, dataloader, split_name):
        dataloaders = {split_name: dataloader}
        return i_g.collect_from_dataloaders(self, dataloaders, [split_name])

    def run(
        self,
        datasets,
        dataloader_creator=None,
        validator=None,
        stat_getter=None,
        validation_interval=1,
        patience=10,
        saver=None,
        resume=None,
        check_initial_accuracy=False,
        **trainer_kwargs,
    ):
        dataloader_creator = c_f.default(dataloader_creator, DataloaderCreator())
        dataloaders = dataloader_creator(**datasets)

        self.dist_init()

        if validator:
            max_epochs = trainer_kwargs.get("max_epochs", 1)
            i_g.add_validation_runner(
                self,
                dataloaders,
                validator,
                stat_getter,
                validation_interval,
                max_epochs,
                saver,
                self.logger,
                check_initial_accuracy,
            )

            self.trainer.add_event_handler(
                Events.EPOCH_STARTED, i_g.early_stopper(patience, validator)
            )

        if saver:
            if not validator:
                self.trainer.add_event_handler(
                    Events.EPOCH_COMPLETED,
                    i_g.save_adapter_without_validator(saver, self.adapter),
                )
            self.trainer.add_event_handler(Events.EPOCH_COMPLETED, saver.save_ignite)

        if resume is not None:
            if resume != "latest":
                raise ValueError("Only 'latest' resume is currently supported")
            if not saver:
                raise ValueError("To resume, a Saver must be provided")
            saver.load_all(
                adapter=self.adapter,
                validator=validator,
                stat_getter=stat_getter,
                framework=self,
                suffix=resume,
            )
            i_g.resume_checks(validator, stat_getter, self)

        if not i_g.is_done(self.trainer, **trainer_kwargs):
            self.trainer.run(dataloaders["train"], **trainer_kwargs)

        if validator:
            return validator.best_score, validator.best_epoch

        return None, None

    def evaluate_best_model(
        self, datasets, validator, saver, epoch, dataloader_creator=None
    ):
        c_f.LOGGER.info("***EVALUATING BEST MODEL***")
        dataloader_creator = c_f.default(dataloader_creator, DataloaderCreator())
        dataloaders = dataloader_creator(**datasets)
        saver.load_adapter(self.adapter, "best")
        collected_data = i_g.collect_from_dataloaders(
            self, dataloaders, validator.required_data
        )
        return i_g.get_validation_score(collected_data, validator, epoch)

    def get_x_collector_step(self, inference, name):
        def collector_step(engine, batch):
            device = idist.device()
            with torch.no_grad():
                batch = c_f.batch_to_device(batch, device)
                features, logits = inference(
                    batch[f"{name}_imgs"], domain=batch[f"{name}_domain"]
                )
            output = {
                "features": features,
                "logits": logits,
                "preds": torch.softmax(logits, dim=1),
                "domain": batch[f"{name}_domain"],
            }
            labels_key = f"{name}_labels"
            if labels_key in batch:
                output["labels"] = batch[labels_key]
            return output

        return collector_step

    def get_labeled_collector_step(self, inference):
        return self.get_x_collector_step(inference, "src")

    def get_unlabeled_collector_step(self, inference):
        return self.get_x_collector_step(inference, "target")

    def get_collector(self, dataset):
        dataset_output = dataset[0]
        return (
            self.labeled_collector
            if len(dataset_output) == 4
            else self.unlabeled_collector
        )

    def set_to_train(self, models):
        def handler(engine):
            c_f.LOGGER.info("Setting models to train() mode")
            models.train()

        return handler

    def set_to_eval(self, models):
        def handler(engine):
            c_f.LOGGER.info("Setting models to eval() mode")
            models.eval()

        return handler



import torch

from .domain_dataset import DomainDataset


class TargetDataset(DomainDataset):
    def __init__(self, dataset, domain=1):
        super().__init__(dataset, domain)

    def __getitem__(self, idx):
        img, _ = self.dataset[idx]
        return {"target_imgs": img, "target_domain": self.domain, "sample_idx": idx}



import torch

from ..layers import MCDLoss
from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .classification import CLossHook
from .features import FeaturesAndLogitsHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import ApplyToListHook, ChainHook, ParallelHook, RepeatHook


class MultipleCLossHook(BaseWrapperHook):
    def __init__(self, num_c=2, loss_fn=None, detach_features=False, **kwargs):
        super().__init__(**kwargs)
        self.num_c = num_c
        loss_fn = c_f.default(loss_fn, torch.nn.CrossEntropyLoss, {"reduction": "none"})
        f_hook = FeaturesAndLogitsHook(domains=["src"], detach_features=detach_features)
        c_hook = CLossHook(loss_fn=loss_fn, detach_features=detach_features)
        c_hook = ApplyToListHook(c_hook, num_c, "_logits$")
        self.hook = ChainHook(f_hook, c_hook)


class MCDLossHook(BaseWrapperHook):
    def __init__(self, detach_features=False, minimize=True, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        self.loss_fn = c_f.default(loss_fn, MCDLoss, {})
        self.hook = FeaturesAndLogitsHook(
            domains=["target"], detach_features=detach_features
        )
        self.minimize = minimize

    def call(self, losses, inputs):
        outputs = self.hook(losses, inputs)[1]
        [target_logits] = c_f.extract(
            [outputs, inputs], c_f.filter(self.hook.out_keys, "_logits$")
        )
        loss = self.loss_fn(*target_logits)
        if not self.minimize:
            loss = -loss
        return {"discrepancy_loss": loss}, outputs

    def _loss_keys(self):
        return ["discrepancy_loss"]


class MCDHook(BaseWrapperHook):
    def __init__(
        self,
        g_opts,
        c_opts,
        discrepancy_loss_fn=None,
        x_weighter=None,
        x_reducer=None,
        y_weighter=None,
        y_reducer=None,
        z_weighter=None,
        z_reducer=None,
        pre_x=None,
        post_x=None,
        pre_y=None,
        post_y=None,
        pre_z=None,
        post_z=None,
        repeat=4,
        **kwargs,
    ):
        super().__init__(**kwargs)
        [pre_x, post_x, pre_y, post_y, pre_z, post_z] = c_f.many_default(
            [pre_x, post_x, pre_y, post_y, pre_z, post_z], [[], [], [], [], [], []]
        )
        x = ChainHook(*pre_x, MultipleCLossHook(), *post_x)
        y = ChainHook(
            *pre_y,
            MultipleCLossHook(detach_features=True),
            MCDLossHook(
                detach_features=True, minimize=False, loss_fn=discrepancy_loss_fn
            ),
            *post_y,
        )
        z = ChainHook(*pre_z, MCDLossHook(loss_fn=discrepancy_loss_fn), *post_z)

        x = OptimizerHook(x, [*c_opts, *g_opts], x_weighter, x_reducer)
        y = OptimizerHook(y, c_opts, y_weighter, y_reducer)
        z = OptimizerHook(z, g_opts, z_weighter, z_reducer)
        s_hook = SummaryHook({"x_loss": x, "y_loss": y, "z_loss": z})
        z = RepeatHook(z, repeat, keep_only_last=True)

        self.hook = ChainHook(ParallelHook(x, y, z), s_hook)



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ...hooks.base import BaseHook
from ...utils import common_functions as c_f
from ...weighters.base_weighter import BaseWeighter


class IgniteEmptyLogger:
    def add_training(self, *args, **kwargs):
        pass

    def add_validation(self, *args, **kwargs):
        pass

    def write(self, *args, **kwargs):
        pass


class IgniteRecordKeeperLogger:
    def __init__(
        self,
        folder=None,
        tensorboard_writer=None,
        record_writer=None,
        attr_list_names=None,
    ):
        import record_keeper

        c_f.LOGGER.info(f"record_keeper version {record_keeper.__version__}")
        from record_keeper import RecordKeeper, RecordWriter
        from torch.utils.tensorboard import SummaryWriter

        tensorboard_writer = c_f.default(
            tensorboard_writer,
            SummaryWriter,
            {"log_dir": folder, "max_queue": 1000000, "flush_secs": 30},
        )
        record_writer = c_f.default(record_writer, RecordWriter, {"folder": folder})
        attr_list_names = c_f.default(
            attr_list_names, pml_cf.list_of_recordable_attributes_list_names()
        )
        self.record_keeper = RecordKeeper(
            tensorboard_writer=tensorboard_writer,
            record_writer=record_writer,
            attributes_to_search_for=attr_list_names,
        )

    def add_training(self, engine):
        adapter = engine.state.adapter
        record_these = [
            ({"engine_output": engine.state.output}, {}),
            (
                adapter.optimizers,
                {
                    "parent_name": "optimizers",
                    "custom_attr_func": optimizer_attr_func,
                },
            ),
            ({"misc": adapter.misc}, {}),
            (
                {"hook": adapter.hook},
                {"recursive_types": [BaseHook, BaseWeighter, torch.nn.Module]},
            ),
        ]
        for record, kwargs in record_these:
            self.record_keeper.update_records(record, engine.state.iteration, **kwargs)

    def add_validation(self, data, epoch):
        self.record_keeper.update_records(data, epoch)

    def write(self, engine):
        self.record_keeper.save_records()
        self.record_keeper.tensorboard_writer.flush()


def optimizer_attr_func(optimizer):
    return {"lr": c_f.get_lr(optimizer)}



from ..layers import AbsLoss
from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .classification import SoftmaxLocallyHook
from .dann import DANNHook, SoftmaxGradientReversalHook
from .domain import DomainLossHook, FeaturesForDomainLossHook
from .features import (
    BaseFeaturesHook,
    FeaturesAndLogitsHook,
    FeaturesChainHook,
    FeaturesHook,
)
from .gan import GANHook
from .utils import ChainHook, MultiplierHook, OnlyNewOutputsHook


class BridgeAndLogitsHook(BaseFeaturesHook):
    def add_if_new(
        self, outputs, full_key, output_vals, inputs, model_name, in_keys, domain
    ):
        # assume model takes 2 inputs: model(x, return_bridge)
        c_f.add_if_new(
            outputs,
            full_key,
            output_vals,
            inputs,
            model_name,
            in_keys,
            other_args=[True],
        )


class GBridgeAndLogitsHook(BridgeAndLogitsHook):
    def __init__(
        self,
        model_name="C",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features"])
        out_suffixes = c_f.default(out_suffixes, ["_logits", "_gbridge"])

        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class DBridgeAndLogitsHook(BridgeAndLogitsHook):
    def __init__(
        self,
        model_name="D",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features_logits"])
        out_suffixes = c_f.default(out_suffixes, ["_dlogits", "_dbridge"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class FeaturesLogitsAndGBridge(FeaturesChainHook):
    def __init__(self, **kwargs):
        hooks = [FeaturesHook(), GBridgeAndLogitsHook()]
        super().__init__(*hooks, **kwargs)


class FeaturesLogitsAndDBridge(BaseWrapperHook):
    def __init__(self, detach_features=False, **kwargs):
        super().__init__(**kwargs)
        self.hook = FeaturesAndLogitsHook(
            detach_features=detach_features,
            detach_logits=detach_features,
            other_hooks=[DBridgeAndLogitsHook()],
        )


class BridgeLossHook(BaseWrapperHook):
    def __init__(self, hook, loss_fn=None, **kwargs):
        super().__init__(**kwargs)
        self.hook = hook
        self.loss_fn = c_f.default(loss_fn, AbsLoss, {})

    def call(self, losses, inputs):
        outputs = self.hook(losses, inputs)[1]
        strs = c_f.filter(self.hook.out_keys, f"_[a-z]bridge$", ["^src", "^target"])
        [src_bridge, target_bridge] = c_f.extract([outputs, inputs], strs)
        return {
            f"src_bridge_loss": self.loss_fn(src_bridge),
            f"target_bridge_loss": self.loss_fn(target_bridge),
        }, outputs

    def _loss_keys(self):
        return [f"src_bridge_loss", f"target_bridge_loss"]


class GBridgeLossHook(BridgeLossHook):
    def __init__(self, hook=None, loss_prefix="g_", **kwargs):
        hook = c_f.default(hook, FeaturesLogitsAndGBridge, {})
        super().__init__(hook=hook, loss_prefix=loss_prefix, **kwargs)


class DBridgeLossHook(BridgeLossHook):
    def __init__(self, hook=None, loss_prefix="d_", detach_features=False, **kwargs):
        hook = c_f.default(
            hook, FeaturesLogitsAndDBridge, {"detach_features": detach_features}
        )
        super().__init__(hook=hook, loss_prefix=loss_prefix, **kwargs)


class GVBHook(DANNHook):
    def __init__(self, pre=None, pre_d=None, pre_g=None, **kwargs):
        # f_hook and d_hook are used inside DomainLossHook
        f_hook = FeaturesForDomainLossHook(use_logits=True)
        d_hook = DBridgeAndLogitsHook()
        apply_to = c_f.filter(f_hook.out_keys, "_logits$")
        gradient_reversal = SoftmaxGradientReversalHook(apply_to)
        [pre, pre_d, pre_g] = c_f.many_default([pre, pre_d, pre_g], [[], [], []])
        pre += [FeaturesLogitsAndGBridge()]
        pre_d += [DBridgeLossHook()]
        pre_g += [GBridgeLossHook()]

        super().__init__(
            pre=pre,
            pre_d=pre_d,
            pre_g=pre_g,
            gradient_reversal=gradient_reversal,
            f_hook=f_hook,
            d_hook=d_hook,
            d_hook_allowed="_dlogits$|_dbridge$",
            **kwargs,
        )


class DBridgeLossWithSoftmaxHook(BaseWrapperHook):
    def __init__(self, detach_features=False, **kwargs):
        super().__init__(**kwargs)
        gbridge_hook = FeaturesLogitsAndGBridge()
        self.hook = ChainHook(
            gbridge_hook,
            SoftmaxLocallyHook(
                c_f.filter(gbridge_hook.out_keys, "_logits"),
                DBridgeLossHook(detach_features=detach_features),
            ),
        )


class GVBGANHook(GANHook):
    def __init__(self, pre=None, pre_d=None, pre_g=None, **kwargs):
        [pre, pre_d, pre_g] = c_f.many_default([pre, pre_d, pre_g], [[], [], []])
        disc_d_hook = DBridgeAndLogitsHook()
        gen_d_hook = DBridgeAndLogitsHook()
        pre_d += [
            FeaturesLogitsAndGBridge(),
            DBridgeLossWithSoftmaxHook(detach_features=True),
        ]
        pre_g += [MultiplierHook(DBridgeLossWithSoftmaxHook(), -1), GBridgeLossHook()]
        super().__init__(
            pre_d=pre_d,
            pre_g=pre_g,
            use_logits=True,
            disc_d_hook=disc_d_hook,
            gen_d_hook=gen_d_hook,
            **kwargs,
        )



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf
from torchmetrics.functional import accuracy

from ..utils import common_functions as c_f


class SufficientAccuracy(torch.nn.Module):
    def __init__(self, threshold, accuracy_func=None, to_probs_func=None):
        super().__init__()
        self.threshold = threshold
        self.accuracy_func = c_f.default(accuracy_func, accuracy)
        self.to_probs_func = c_f.default(to_probs_func, torch.nn.Sigmoid())
        pml_cf.add_to_recordable_attributes(
            self, list_of_names=["accuracy", "threshold"]
        )

    def forward(self, x, labels):
        with torch.no_grad():
            x = self.to_probs_func(x)
            labels = labels.type(torch.int)
            self.accuracy = self.accuracy_func(x, labels).item()
        return self.accuracy > self.threshold

    def extra_repr(self):
        return c_f.extra_repr(self, ["threshold"])



import logging

import numpy as np
import torch
import torch.nn.functional as F
from pytorch_metric_learning.utils import common_functions as pml_cf
from sklearn.model_selection import train_test_split

from ..adapters import Finetuner
from ..containers import Models, Optimizers
from ..datasets import SourceDataset
from ..frameworks import Ignite
from ..models import Discriminator
from ..utils import common_functions as c_f
from .accuracy_validator import AccuracyValidator
from .base_validator import BaseValidator


class DeepEmbeddedValidator(BaseValidator):
    def __init__(self, layer="features", **kwargs):
        super().__init__(**kwargs)
        self.layer = layer

    def compute_score(self, src_train, src_val, target_train):
        init_logging_level = c_f.LOGGER.level
        c_f.LOGGER.setLevel(logging.WARNING)
        weights = get_weights(
            src_train[self.layer], src_val[self.layer], target_train[self.layer]
        )
        error_per_sample = F.cross_entropy(
            src_val["logits"], src_val["labels"], reduction="none"
        )
        output = get_dev_risk(weights, error_per_sample[:, None])
        c_f.LOGGER.setLevel(init_logging_level)
        return output

    @property
    def maximize(self):
        return False


#########################################################################
#### ADAPTED FROM https://github.com/thuml/Deep-Embedded-Validation #####
#########################################################################


def get_dev_risk(weight, error):
    """
    :param weight: shape [N, 1], the importance weight for N source samples in the validation set
    :param error: shape [N, 1], the error value for each source sample in the validation set
    (typically 0 for correct classification and 1 for wrong classification)
    """
    if torch.any(weight < 0) or torch.any(error < 0):
        raise ValueError("weights and errors must be positive")

    weight = pml_cf.to_numpy(weight)
    error = pml_cf.to_numpy(error)

    N, d = weight.shape
    _N, _d = error.shape
    assert N == _N and d == _d, "dimension mismatch!"
    weighted_error = weight * error
    cov = np.cov(np.concatenate((weighted_error, weight), axis=1), rowvar=False)[0][1]
    var_w = np.var(weight, ddof=1)
    eta = -cov / var_w
    return np.mean(weighted_error) + eta * np.mean(weight) - eta


def get_weights(source_feature, validation_feature, target_feature):
    """
    :param source_feature: shape [N_tr, d], features from training set
    :param validation_feature: shape [N_v, d], features from validation set
    :param target_feature: shape [N_te, d], features from test set
    :return:
    """

    device = source_feature.device
    source_feature = pml_cf.to_numpy(source_feature)
    validation_feature = pml_cf.to_numpy(validation_feature)
    target_feature = pml_cf.to_numpy(target_feature)

    N_s, d = source_feature.shape
    N_t, _d = target_feature.shape
    source_feature = source_feature.copy()
    target_feature = target_feature.copy()
    all_feature = np.concatenate((source_feature, target_feature))
    all_label = np.asarray([1] * N_s + [0] * N_t, dtype=np.int64)
    (
        feature_for_train,
        feature_for_test,
        label_for_train,
        label_for_test,
    ) = train_test_split(all_feature, all_label, train_size=0.8)

    train_set = SourceDataset(
        pml_cf.EmbeddingDataset(feature_for_train, label_for_train)
    )
    val_set = SourceDataset(pml_cf.EmbeddingDataset(feature_for_test, label_for_test))

    decays = [1e-1, 3e-2, 1e-2, 3e-3, 1e-3, 3e-4, 1e-4, 3e-5, 1e-5]
    val_acc = []
    trainers = []
    epochs = 2

    for decay in decays:
        models = Models(
            {
                "G": torch.nn.Identity(),
                "C": Discriminator(d, h=d, out_size=2).to(device),
            }
        )
        optimizers = Optimizers(
            (torch.optim.Adam, {"lr": 0.001, "weight_decay": decay})
        )
        trainer = Finetuner(models=models, optimizers=optimizers)
        trainer = Ignite(trainer, with_pbars=False)
        datasets = {"train": train_set, "src_val": val_set}
        acc, _ = trainer.run(
            datasets,
            max_epochs=epochs,
            validator=AccuracyValidator(),
            validation_interval=epochs,
        )
        val_acc.append(acc)
        trainers.append(trainer)

    index = val_acc.index(max(val_acc))

    validation_set = SourceDataset(
        pml_cf.EmbeddingDataset(validation_feature, np.ones(len(validation_feature)))
    )
    trainer = trainers[index]
    dataloader = torch.utils.data.DataLoader(validation_set)
    domain_out = trainer.get_all_outputs(dataloader, "val")
    domain_out = domain_out["val"]["preds"]
    return (domain_out[:, :1] / domain_out[:, 1:]) * (float(N_s) / N_t)



import torch

from ..containers import KeyEnforcer
from ..hooks import CDANHook, DomainConfusionHook, GANHook, VADAHook
from ..utils import common_functions as c_f
from .base_adapter import BaseGCDAdapter


class GAN(BaseGCDAdapter):
    hook_cls = GANHook

    def init_hook(self, hook_kwargs):
        g_opts = c_f.extract(self.optimizers, ["G", "C"])
        d_opts = [self.optimizers["D"]]
        self.hook = self.hook_cls(d_opts=d_opts, g_opts=g_opts, **hook_kwargs)


class CDAN(GAN):
    hook_cls = CDANHook

    def get_key_enforcer(self):
        ke = super().get_key_enforcer()
        ke.requirements["misc"] = ["feature_combiner"]
        return ke


class DomainConfusion(GAN):
    hook_cls = DomainConfusionHook


class VADA(GAN):
    hook_cls = VADAHook

    def init_containers_and_check_keys(self):
        models = self.containers["models"]
        misc = self.containers["misc"]
        misc["combined_model"] = torch.nn.Sequential(models["G"], models["C"])
        super().init_containers_and_check_keys()

    def get_key_enforcer(self):
        ke = super().get_key_enforcer()
        ke.requirements["misc"] = ["combined_model"]
        return ke



import torch

from ..utils import common_functions as c_f
from .entropy_loss import entropy_after_softmax
from .entropy_weights import EntropyWeights
from .normalizers import SumNormalizer


# reference https://github.com/thuml/Versatile-Domain-Adaptation
class MCCLoss(torch.nn.Module):
    def __init__(self, T=1, entropy_weighter=None):
        super().__init__()
        self.T = T
        self.entropy_weighter = c_f.default(
            entropy_weighter,
            EntropyWeights(
                after_softmax=True, normalizer=SumNormalizer(scale_by_batch_size=True)
            ),
        )

    def forward(self, x):
        Y = torch.nn.functional.softmax(x / self.T, dim=1)
        H_weights = self.entropy_weighter(Y.detach())
        C = torch.linalg.multi_dot([Y.t(), torch.diag(H_weights), Y])
        C = C / torch.sum(C, dim=1)
        return (torch.sum(C) - torch.trace(C)) / C.shape[0]

    def extra_repr(self):
        return c_f.extra_repr(self, ["T"])



import inspect
import json
import os
from abc import ABC, abstractmethod

import numpy as np
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


class BaseValidator(ABC):
    def __init__(self, normalizer=None, key_map=None, ignore_epoch=0):
        self.normalizer = c_f.default(normalizer, return_raw)
        self.score_history = np.array([])
        self.raw_score_history = np.array([])
        self.epochs = np.array([], dtype=int)
        self.key_map = c_f.default(key_map, {})
        self.ignore_epoch = ignore_epoch
        pml_cf.add_to_recordable_attributes(
            self,
            list_of_names=["latest_score", "best_score", "latest_epoch", "best_epoch"],
        )

    def _required_data(self):
        args = inspect.getfullargspec(self.compute_score).args
        args.remove("self")
        return args

    @property
    def required_data(self):
        output = set(self._required_data()) - set(self.key_map.values())
        output = list(output)
        for k, v in self.key_map.items():
            output.append(k)
        return output

    @abstractmethod
    def compute_score(self):
        pass

    def score(self, epoch, **kwargs):
        kwargs = self.kwargs_check(epoch, kwargs)
        score = self.compute_score(**kwargs)
        self.append_to_history_and_normalize(score, epoch)
        return self.latest_score

    def kwargs_check(self, epoch, kwargs):
        if epoch in self.epochs:
            raise ValueError(f"Epoch {epoch} has already been evaluated")
        if kwargs.keys() != set(self.required_data):
            raise ValueError(
                f"Input to compute_score has keys = {kwargs.keys()} but should have keys {self.required_data}"
            )
        return c_f.map_keys(kwargs, self.key_map)

    def append_to_history_and_normalize(self, score, epoch):
        self.raw_score_history = np.append(self.raw_score_history, score)
        self.epochs = np.append(self.epochs, epoch)
        self.score_history = self.normalizer(self.raw_score_history)

    @property
    def score_history_ignore_epoch(self):
        return remove_ignore_epoch(self.score_history, self.epochs, self.ignore_epoch)

    @property
    def epochs_ignore_epoch(self):
        return remove_ignore_epoch(self.epochs, self.epochs, self.ignore_epoch)

    def has_valid_history(self, ignore_ignore_epoch=True):
        x = (
            self.score_history_ignore_epoch
            if ignore_ignore_epoch
            else self.score_history
        )
        return len(x) > 0 and np.isfinite(x).any()

    @property
    def best_score(self):
        if self.has_valid_history():
            return self.score_history_ignore_epoch[self.best_idx]

    @property
    def best_epoch(self):
        if self.has_valid_history():
            return self.epochs_ignore_epoch[self.best_idx]

    @property
    def best_idx(self):
        if self.has_valid_history():
            return (
                np.nanargmax(self.score_history_ignore_epoch)
                if self.maximize
                else np.nanargmin(self.score_history_ignore_epoch)
            )

    @property
    def latest_epoch(self):
        if self.has_valid_history(False):
            return self.epochs[-1]

    @property
    def latest_score(self):
        if self.has_valid_history(False):
            return self.score_history[-1]

    @property
    def latest_is_best(self):
        if self.has_valid_history(False):
            return self.best_epoch == self.latest_epoch
        return False

    @property
    def maximize(self):
        return True

    def __repr__(self):
        return c_f.nice_repr(self, self.extra_repr(), {})

    def extra_repr(self):
        return c_f.extra_repr(
            self, ["latest_score", "best_score", "best_epoch", "required_data"]
        )


def score_dict_to_str(d):
    return "\n\t" + "\n\t".join([f"{k}: {v}" for k, v in d.items()])


def remove_ignore_epoch(x, epochs, ignore_epoch):
    if ignore_epoch is not None:
        return x[epochs != ignore_epoch]
    return x


def return_raw(raw_score_history):
    return raw_score_history



import torch

from ..utils import common_functions as c_f


class SlicedWasserstein(torch.nn.Module):
    def __init__(self, m=128):
        super().__init__()
        self.m = 128

    def forward(self, x, y):
        d = x.shape[1]
        proj = torch.randn(d, self.m, device=x.device)
        proj = torch.nn.functional.normalize(proj, dim=0)
        x = torch.matmul(x, proj)
        y = torch.matmul(y, proj)
        x, _ = torch.sort(x, dim=0)
        y, _ = torch.sort(y, dim=0)
        return torch.mean((x - y) ** 2)

    def extra_repr(self):
        return c_f.extra_repr(self, ["m"])



import os

from .base_dataset import BaseDataset
from .utils import check_length


class MNISTM(BaseDataset):
    def __init__(self, root, train, transform, **kwargs):
        super().__init__(domain="MNISTM", **kwargs)
        if not isinstance(train, bool):
            raise TypeError("train should be True or False")
        name = "train" if train else "test"
        labels_file = os.path.join(root, "mnist_m", f"mnist_m_{name}_labels.txt")
        img_dir = os.path.join(root, "mnist_m", f"mnist_m_{name}")

        with open(labels_file) as f:
            content = [line.rstrip().split(" ") for line in f]
        self.img_paths = [os.path.join(img_dir, x[0]) for x in content]
        check_length(self, {"train": 59001, "test": 9001}[name])
        self.labels = [int(x[1]) for x in content]
        self.transform = transform



import torch

from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .features import FeaturesChainHook, FeaturesHook, LogitsHook
from .utils import ParallelHook


class DomainSpecificFeaturesHook(FeaturesHook):
    def add_if_new(
        self, outputs, full_key, output_vals, inputs, model_name, in_keys, domain
    ):
        [domain] = c_f.extract(inputs, [f"{domain}_domain"])
        c_f.add_if_new(
            outputs,
            full_key,
            output_vals,
            inputs,
            model_name,
            in_keys,
            other_args=[domain],
        )


class DomainSpecificLogitsHook(LogitsHook, DomainSpecificFeaturesHook):
    pass


class AdaBNHook(BaseWrapperHook):
    def __init__(self, domains=None, **kwargs):
        super().__init__(**kwargs)
        domains = c_f.default(domains, ["src", "target"])
        hooks = []
        for d in domains:
            f_hook = DomainSpecificFeaturesHook(domains=[d], detach=True)
            l_hook = DomainSpecificLogitsHook(domains=[d], detach=True)
            hooks.append(FeaturesChainHook(f_hook, l_hook))
        self.hook = ParallelHook(*hooks)

    def call(self, losses, inputs):
        with torch.no_grad():
            losses, outputs = self.hook(losses, inputs)
        return losses, outputs



from ..containers import KeyEnforcer
from ..hooks import DANNHook, GVBHook
from ..layers import ModelWithBridge
from .base_adapter import BaseGCDAdapter


class DANN(BaseGCDAdapter):
    hook_cls = DANNHook

    def init_hook(self, hook_kwargs):
        self.hook = self.hook_cls(opts=list(self.optimizers.values()), **hook_kwargs)


class GVB(DANN):
    hook_cls = GVBHook

    def init_containers_and_check_keys(self):
        models = self.containers["models"]
        for k in ["D", "C"]:
            if not isinstance(models[k], ModelWithBridge):
                models[k] = ModelWithBridge(models[k])
        super().init_containers_and_check_keys()



import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f


class GradientReversal(torch.nn.Module):
    def __init__(self, weight=1.0):
        super().__init__()
        self.register_buffer("weight", torch.tensor([weight]))
        pml_cf.add_to_recordable_attributes(self, "weight")

    def update_weight(self, new_weight):
        self.weight[0] = new_weight

    def forward(self, x):
        return _GradientReversal.apply(x, pml_cf.to_device(self.weight, x))

    def extra_repr(self):
        return c_f.extra_repr(self, ["weight"])


class _GradientReversal(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input, weight):
        ctx.weight = weight
        return input

    @staticmethod
    def backward(ctx, grad_output):
        return -ctx.weight * grad_output, None



import errno
import glob
import itertools
import json
import logging
import os
import re
import shutil
from collections import OrderedDict

import numpy as np
import torch
from pytorch_metric_learning.utils import common_functions as pml_cf

LOGGER_NAME = "PDA"
LOGGER = logging.getLogger(LOGGER_NAME)


def set_logger_name(name):
    global LOGGER_NAME
    global LOGGER
    LOGGER_NAME = name
    LOGGER = logging.getLogger(LOGGER_NAME)


def has_no_parameters(model):
    return len(list(model.parameters())) == 0


def is_optimizer(optimizer):
    return isinstance(optimizer, torch.optim.Optimizer)


def is_none(x):
    return x is None


def default(x, default_x, kwargs=None, condition=None):
    if condition is None:
        condition = is_none
    if kwargs is not None:
        if isinstance(kwargs, dict):
            return default_x(**kwargs) if condition(x) else x
        if isinstance(kwargs, list):
            return default_x(*kwargs) if condition(x) else x
    return default_x if condition(x) else x


def many_default(x, default_x, kwargs=None, condition=None):
    return [default(x[i], default_x[i], kwargs, condition) for i in range(len(x))]


def dict_pop_lazy(x, key, *args, **kwargs):
    y = x.pop(key, None)
    return default(y, *args, **kwargs)


def add_if_new(d, key, x, kwargs, model_name, in_keys, other_args=None):
    # if key is list then assume model returns multiple args
    if not is_list_or_tuple(key) or not is_list_or_tuple(x):
        raise TypeError("key and x must both be a list or tuple")
    condition = is_none
    if any(condition(y) for y in x):
        model = kwargs[model_name]
        input_vals = [kwargs[k] for k in in_keys]
        if other_args is not None:
            input_vals += other_args
        new_x = default(None, model, input_vals, is_none)
        if len(x) > 1:
            if not is_list_or_tuple(new_x) or len(new_x) != len(x):
                raise TypeError(
                    "if input x and key are lists, then output of model must be a list of the same length"
                )
            for i in range(len(x)):
                if condition(x[i]):
                    d[key[i]] = new_x[i]
        else:
            d[key[0]] = new_x


def class_default(cls, x, default):
    return x(cls) if x else default


def list_diff(x, y):
    return list(set(x) - set(y))


def is_list_or_tuple(x):
    return isinstance(x, (list, tuple))


def batch_to_device(batch, device):
    if is_list_or_tuple(batch):
        return [pml_cf.to_device(x, device=device) for x in batch]
    if isinstance(batch, dict):
        return {k: pml_cf.to_device(x, device=device) for k, x in batch.items()}
    return pml_cf.to_device(batch, device=device)


def first_key(in_dict):
    return list(in_dict.keys())[0]


def first_val(in_dict):
    return in_dict[first_key(in_dict)]


def kronecker_product(x, y):
    batch_size = x.shape[0]
    tensor_prod = torch.bmm(x.unsqueeze(2), y.unsqueeze(1))
    return tensor_prod.view(batch_size, -1)


def makedir_if_not_there(dir_name):
    try:
        os.makedirs(dir_name)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise


def filter_kwargs(kwargs, keep):
    return {x: kwargs[x] for x in keep}


def val_dataloader_checks(dataloader):
    if dataloader.drop_last:
        raise ValueError("drop_last should be False when doing validation")
    if not isinstance(dataloader.sampler, torch.utils.data.SequentialSampler):
        raise ValueError("shuffle should be False when doing validation")


def val_collected_data_checks(collected, dataset):
    for k, v in collected.items():
        if len(v) != len(dataset):
            raise ValueError(
                f"The length of {k} should be equal to the length of the validation dataset"
            )


def add_prefix(x, prefix=""):
    prefix = f"{prefix}_" if prefix != "" else prefix
    return f"{prefix}{x}"


def add_suffix(x, suffix=""):
    suffix = f"_{suffix}" if suffix != "" else suffix
    return f"{x}{suffix}"


def class_as_prefix(obj, x, prefix="", suffix=""):
    x = add_prefix(x, cls_name(obj))
    x = add_prefix(x, prefix)
    x = add_suffix(x, suffix)
    return x


def full_path(folder, filename):
    return os.path.join(folder, filename)


def has_state_dict(x):
    return hasattr(x, "state_dict")


def copy_file(src, dst):
    shutil.copyfile(src, dst)


def save_torch_module(x, folder, filename):
    if isinstance(x, torch.nn.DataParallel):
        x = x.module
    makedir_if_not_there(folder)
    filename = full_path(folder, filename)
    LOGGER.debug(f"Saving {filename}")
    pml_cf.save_model(x, filename)


def load_torch_module(x, folder, filename):
    if isinstance(x, torch.nn.DataParallel):
        x = x.module
    filename = full_path(folder, filename)
    LOGGER.debug(f"Loading {filename}")
    pml_cf.load_model(x, filename, torch.device("cuda"))


def save_json(x, folder, filename):
    makedir_if_not_there(folder)
    filename = full_path(folder, filename)
    with open(filename, "w") as fp:
        LOGGER.debug(f"Saving {filename}")
        json.dump(x, fp)


def load_json(folder, filename):
    filename = full_path(folder, filename)
    with open(filename, "r") as fp:
        LOGGER.debug(f"Loading {filename}")
        data = json.load(fp)
    return data


def save_npy(x, folder, filename):
    makedir_if_not_there(folder)
    filename = full_path(folder, filename)
    LOGGER.debug(f"Saving {filename}")
    np.save(filename, x, allow_pickle=False)


def load_npy(folder, filename):
    filename = full_path(folder, filename)
    LOGGER.debug(f"Loading {filename}")
    return np.load(filename, allow_pickle=False)


def enumerate_to_dict(x):
    if isinstance(x, list):
        return {str(i): v for i, v in enumerate(x)}
    elif isinstance(x, dict):
        return x
    else:
        raise TypeError("input must be a list or dict")


def delete_all_but(folder, basename, extension, to_keep):
    basenames_to_keep = []
    for prefix, suffix in to_keep:
        new_basename = add_suffix(add_prefix(basename, prefix), suffix)
        basenames_to_keep.append(new_basename)
    files_to_keep = [full_path(folder, f"{x}{extension}") for x in basenames_to_keep]
    all_matching_files = glob.glob(full_path(folder, f"*{basename}*{extension}"))

    for f in all_matching_files:
        if f not in files_to_keep:
            LOGGER.debug(f"Deleting {f}")
            os.remove(f)


def state_dicts_are_equal(x, y, rtol=None):
    if x.keys() != y.keys():
        return False
    for k, v1 in x.items():
        v2 = y[k]
        if type(v1) != type(v2):
            return False
        if torch.is_tensor(v1):
            if rtol is None and not torch.equal(v1, v2):
                return False
            elif rtol and not torch.allclose(v1, v2, rtol=rtol):
                return False
        elif isinstance(v1, dict) and not state_dicts_are_equal(v1, v2):
            return False
    return True


def set_layers_mode(mode, layer_types=None, layer_names=None):
    if layer_types is None and layer_names is None:
        raise ValueError("layer_types and layer_names cannot both be None")
    if mode not in ["train", "eval"]:
        raise ValueError("mode must be either 'train' or 'eval'")

    def helper(m, mode):
        m.eval() if mode == "eval" else m.train()

    def set_to_mode(m):
        if layer_names is not None:
            classname = cls_name(m)
            if any(L in classname for L in layer_names):
                helper(m, mode)
        if layer_types is not None:
            if isinstance(m, layer_types):
                helper(m, mode)

    return set_to_mode


def batchnorm_types():
    return (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d)


def get_lr(optimizer):
    return optimizer.param_groups[0]["lr"]


def cls_name(x):
    return x.__class__.__name__


def assert_keys_are_present_cls(cls, attr_name, x):
    assert_keys_are_present(getattr(cls, attr_name), attr_name, x)


def assert_keys_are_present(attr, attr_name, x):
    if any(k not in x for k in attr.keys()):
        raise KeyError(f"{attr_name} ({attr}) has keys that are not in input ({x}).")


def get_or_pop(x, k, pop):
    return x.pop(k) if pop else x[k]


def extract(x, keys, pop=False):
    if isinstance(x, list):
        output = []
        for k in keys:
            success = False
            for y in x:
                try:
                    output.append(get_or_pop(y, k, pop))
                    success = True
                    break
                except KeyError:
                    continue
            if not success:
                raise KeyError(f"{k} not found")
        return output
    else:
        return [get_or_pop(x, k, pop) for k in keys]


def filter(x, regex_str, return_order=None):
    r = re.compile(regex_str)
    x = [y for y in x if r.search(y)]
    if return_order:
        output = []
        for s in return_order:
            r = re.compile(s)
            output.extend([y for y in x if r.search(y)])
        return output
    return x


def to_dict(x, keys):
    return {k: x[i] for i, k in enumerate(keys)}


def concat_tensor_dict(x, keys, dim):
    return torch.cat(extract(x, keys), dim=dim)


def zero_loss():
    return torch.tensor(0, dtype=torch.float, requires_grad=True)


def zero_back_step(loss, optimizers):
    for x in optimizers:
        x.zero_grad()
    if loss.grad_fn is not None:
        loss.backward()
        for x in optimizers:
            x.step()


def len_one_tensor(x):
    return x.dim() == 0 or len(x) == 1


def reinit_layer(layer):
    if hasattr(layer, "reset_parameters"):
        layer.reset_parameters()


def reinit(model):
    for L in model.children():
        try:
            for layer in L:
                reinit_layer(layer)
        except TypeError:
            reinit_layer(L)
    return model


def map_keys(x, key_map, must_be_subset=True):
    if must_be_subset and not set(key_map.keys()).issubset(set(x.keys())):
        raise KeyError("key_map keys must be a subset of the input")
    output = {}
    for k, v in x.items():
        new_k = key_map.get(k, k)
        if new_k != k or new_k not in output:
            output[new_k] = v
    return output


# https://stackoverflow.com/a/6117124
def map_keys_substrings(x, key_map):
    rep = dict((re.escape(k), v) for k, v in key_map.items())
    pattern = re.compile("|".join(rep.keys()))
    return pattern.sub(lambda m: rep[re.escape(m.group(0))], x)


def join_lists(x):
    return list(itertools.chain(*x))


def attrs_of_type(cls, obj):
    attrs = [a for a in dir(cls) if not a.startswith("__")]
    attrs = {h: getattr(cls, h) for h in attrs}
    return {k: v for k, v in attrs.items() if isinstance(v, obj)}


def append_error_message(e, msg):
    if len(e.args) >= 1:
        e.args = (e.args[0] + msg,) + e.args[1:]


def requires_grad(x, does=True):
    if isinstance(x, (list, tuple, set)):
        return all(requires_grad(y, does) for y in x)
    if isinstance(x, (dict)):
        return all(requires_grad(y, does) for y in x.values())
    x = x.requires_grad
    return x if does else not x


def to_set(x):
    if isinstance(x, list):
        if all(isinstance(y, dict) for y in x):
            return set().union(*x)
        return set(x)
    return set(x)


def check_domain(cls, domain, keep_len=False):
    if domain is not None:
        if len(torch.unique(domain)) > 1:
            raise ValueError(
                f"{c_f.cls_name(cls)} inference only supports one domain per batch"
            )
        if not keep_len and not len_one_tensor(domain):
            domain = domain[0]
    return domain


def extra_repr(cls, attr_names, delimiter="\n"):
    return delimiter.join([f"{a}={str(getattr(cls, a))}" for a in attr_names])


# copied from https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/module.py
def add_indent(s_, num_spaces=2, not_first=False):
    s = s_.split("\n")
    # don't do anything for single-line stuff
    if len(s) == 1:
        return s_
    if not_first:
        first = s.pop(0)
    s = [(num_spaces * " ") + line for line in s]
    s = "\n".join(s)
    if not_first:
        s = first + "\n" + s
    return s


# copied from https://github.com/pytorch/pytorch/blob/master/torch/nn/modules/module.py
def nice_repr(cls, extra_repr, children):
    # We treat the extra repr like the sub-module, one item per line
    extra_lines = []
    # empty string will be split into list ['']
    if extra_repr:
        extra_lines = extra_repr.split("\n")
    child_lines = []
    for key, value in children.items():
        mod_str = repr(value)
        mod_str = add_indent(mod_str, 2, not_first=True)
        child_lines.append("(" + key + "): " + mod_str)
    lines = extra_lines + child_lines

    main_str = cls_name(cls) + "("
    if lines:
        # simple one-liner info, which most builtin Modules will use
        if len(extra_lines) == 1 and not child_lines:
            main_str += extra_lines[0]
        else:
            main_str += "\n  " + "\n  ".join(lines) + "\n"

    main_str += ")"
    return main_str



import torch

from ..utils import common_functions as c_f


class DataloaderCreator:
    def __init__(
        self,
        train_kwargs=None,
        val_kwargs=None,
        train_names=None,
        val_names=None,
        all_train=False,
        all_val=False,
        batch_size=32,
        num_workers=0,
    ):
        self.train_kwargs = c_f.default(
            train_kwargs,
            {
                "batch_size": batch_size,
                "num_workers": num_workers,
                "shuffle": True,
                "drop_last": True,
            },
        )
        self.val_kwargs = c_f.default(
            val_kwargs,
            {
                "batch_size": batch_size,
                "num_workers": num_workers,
                "shuffle": False,
                "drop_last": False,
            },
        )
        self.train_names = c_f.default(train_names, ["train"])
        self.val_names = c_f.default(
            val_names, ["src_train", "target_train", "src_val", "target_val"]
        )
        if not set(self.train_names).isdisjoint(self.val_names):
            raise ValueError(
                f"train_names {self.train_names} must be disjoint from val_names {self.val_names}"
            )
        if all_train and all_val:
            raise ValueError("all_train and all_val cannot both be True")
        self.all_train = all_train
        self.all_val = all_val

    def __call__(self, **kwargs):
        output = {}
        for k, v in kwargs.items():
            if self.all_train:
                dataloader_kwargs = self.train_kwargs
            elif self.all_val:
                dataloader_kwargs = self.val_kwargs
            elif k in self.train_names:
                dataloader_kwargs = self.train_kwargs
            elif k in self.val_names:
                dataloader_kwargs = self.val_kwargs
            else:
                raise ValueError(
                    f"Dataset split name must be in {self.train_names} or {self.val_names}, or one of self.all_train or self.all_val must be true"
                )
            output[k] = torch.utils.data.DataLoader(v, **dataloader_kwargs)
        return output



import torch

from .entropy_loss import entropy


class DiversityLoss(torch.nn.Module):
    def forward(self, logits):
        logits = torch.mean(logits, dim=0, keepdim=True)
        return -torch.mean(entropy(logits))



import torch

from ..utils import common_functions as c_f


def sum_normalizer(x, detach=False, scale_by_batch_size=False):
    y = torch.sum(x)
    if detach:
        y = y.detach()
    if scale_by_batch_size:
        x = x * (x.shape[0])
    return x / y


def min_max_normalizer(x, detach=False):
    x_min = torch.min(x)
    x_max = torch.max(x)
    if detach:
        x_min = x_min.detach()
        x_max = x_max.detach()
    return (x - x_min) / (x_max - x_min)


def no_normalizer(x):
    return x


class SumNormalizer(torch.nn.Module):
    def __init__(self, detach=False, scale_by_batch_size=False):
        super().__init__()
        self.detach = detach
        self.scale_by_batch_size = scale_by_batch_size

    def forward(self, x):
        return sum_normalizer(
            x, detach=self.detach, scale_by_batch_size=self.scale_by_batch_size
        )

    def extra_repr(self):
        return c_f.extra_repr(self, ["detach", "scale_by_batch_size"])


class MinMaxNormalizer(torch.nn.Module):
    def __init__(self, detach=False):
        super().__init__()
        self.detach = detach

    def forward(self, x):
        return min_max_normalizer(x, detach=self.detach)

    def extra_repr(self):
        return c_f.extra_repr(self, ["detach"])


class NoNormalizer(torch.nn.Module):
    def forward(self, x):
        return no_normalizer(x)



import torch


def covariance(x):
    batch_size = x.shape[0]
    mm1 = torch.mm(x.t(), x)
    cols_summed = torch.sum(x, dim=0)
    mm2 = torch.mm(cols_summed.unsqueeze(1), cols_summed.unsqueeze(0))
    return (1.0 / (batch_size - 1)) * (mm1 - (1.0 / batch_size) * mm2)


class CORALLoss(torch.nn.Module):
    def forward(self, x, y):
        embedding_size = x.shape[1]
        cx = covariance(x)
        cy = covariance(y)
        squared_fro_norm = torch.linalg.norm(cx - cy, ord="fro") ** 2
        return squared_fro_norm / (4 * (embedding_size ** 2))



import torch
import torch.nn.functional as F
from pytorch_metric_learning.utils.accuracy_calculator import AccuracyCalculator

from .base_validator import BaseValidator


class KNNValidator(BaseValidator):
    def __init__(
        self,
        layer="features",
        l2_normalize=False,
        metric="precision_at_1",
        k=1,
        **kwargs
    ):
        super().__init__(**kwargs)
        self.acc_fn = AccuracyCalculator(k=k, include=(metric,), avg_of_avgs=True)
        self.layer = layer
        self.l2_normalize = l2_normalize
        self.metric = metric
        self.k = k

    def compute_score(self, src_train, target_train):
        features = torch.cat([src_train[self.layer], target_train[self.layer]], dim=0)
        if self.l2_normalize:
            features = F.normalize(features, dim=1)
        labels = torch.cat([src_train["domain"], target_train["domain"]], dim=0)
        accuracies = self.acc_fn.get_accuracy(features, features, labels, labels, True)
        return accuracies[self.metric]

    @property
    def maximize(self):
        return False


class ClusterValidator(KNNValidator):
    def __init__(self, layer="features", metric="AMI", **kwargs):
        if metric not in ["AMI", "NMI"]:
            raise ValueError("metric must be 'AMI' or 'NMI'")
        super().__init__(layer=layer, metric=metric, **kwargs)



from ..layers import EntropyLoss
from .base_validator import BaseValidator


class EntropyValidator(BaseValidator):
    def compute_score(self, target_train):
        return EntropyLoss()(target_train["logits"]).item()

    @property
    def maximize(self):
        return False



import torch

from ..datasets import (
    CombinedSourceAndTargetDataset,
    DataloaderCreator,
    PseudoLabeledDataset,
    TargetDataset,
)
from ..utils import common_functions as c_f


def get_predicted_labels(adapter, datasets, dataset_name, dataloader_creator):
    dataloader = dataloader_creator(**{dataset_name: datasets[dataset_name]})
    dataloader = dataloader[dataset_name]
    c_f.val_dataloader_checks(dataloader)
    logits = adapter.get_all_outputs(dataloader, dataset_name)[dataset_name]["logits"]
    return torch.argmax(logits, dim=1).cpu().numpy()


def get_pseudo_labeled_dataset(adapter, datasets, dataset_name, dataloader_creator):
    pseudo_labels = get_predicted_labels(
        adapter, datasets, dataset_name, dataloader_creator
    )
    return PseudoLabeledDataset(datasets[dataset_name].dataset, pseudo_labels)


class ReverseValidator:
    def __init__(self):
        self.pseudo_train = None
        self.pseudo_val = None

    def run(
        self,
        forward_adapter,
        reverse_adapter,
        forward_kwargs,
        reverse_kwargs,
        pl_dataloader_creator=None,
    ):
        if "datasets" in reverse_kwargs:
            raise KeyError(
                "'datasets' should not be in reverse_kwargs because the reverse datasets will be pseudo labeled."
            )
        if "validator" not in reverse_kwargs:
            raise KeyError("reverse_kwargs must include 'validator'")

        forward_adapter.run(**forward_kwargs)
        if all(x in forward_kwargs for x in ["validator", "saver"]):
            forward_kwargs["saver"].load_adapter(forward_adapter.adapter, "best")

        datasets = forward_kwargs["datasets"]
        pl_dataloader_creator = c_f.default(
            pl_dataloader_creator, DataloaderCreator, {"all_val": True}
        )

        d = {}
        d["src_train"] = get_pseudo_labeled_dataset(
            forward_adapter, datasets, "target_train", pl_dataloader_creator
        )
        d["src_val"] = get_pseudo_labeled_dataset(
            forward_adapter, datasets, "target_val", pl_dataloader_creator
        )
        d["target_train"] = TargetDataset(datasets["src_train"].dataset)
        d["target_val"] = TargetDataset(datasets["src_val"].dataset)
        d["train"] = CombinedSourceAndTargetDataset(d["src_train"], d["target_train"])

        self.pseudo_train = d["src_train"]
        self.pseudo_val = d["src_val"]

        reverse_kwargs["datasets"] = d
        return reverse_adapter.run(**reverse_kwargs)



from ..utils import common_functions as c_f
from .accuracy_validator import AccuracyValidator


class ErrorValidator(AccuracyValidator):
    def compute_score(self, src_val):
        return 1 - super().compute_score(src_val)

    @property
    def maximize(self):
        return False



from ..utils import common_functions as c_f
from .base import BaseWrapperHook
from .classification import CLossHook
from .domain import DomainLossHook, FeaturesForDomainLossHook
from .optimizer import OptimizerHook, SummaryHook
from .utils import AssertHook, ChainHook, OnlyNewOutputsHook


class GANHook(BaseWrapperHook):
    def __init__(
        self,
        d_opts,
        g_opts,
        d_weighter=None,
        d_reducer=None,
        g_weighter=None,
        g_reducer=None,
        pre_d=None,
        post_d=None,
        pre_g=None,
        post_g=None,
        use_logits=False,
        disc_hook=None,
        gen_hook=None,
        disc_f_hook=None,
        gen_f_hook=None,
        disc_d_hook=None,
        gen_d_hook=None,
        c_hook=None,
        disc_conditions=None,
        disc_alts=None,
        gen_conditions=None,
        gen_alts=None,
        disc_domains=None,
        gen_domains=None,
        disc_domain_loss_fn=None,
        gen_domain_loss_fn=None,
        **kwargs
    ):
        super().__init__(**kwargs)
        [pre_d, post_d, pre_g, post_g] = c_f.many_default(
            [pre_d, post_d, pre_g, post_g], [[], [], [], []]
        )
        disc_f_hook = c_f.default(
            disc_f_hook,
            FeaturesForDomainLossHook,
            {"detach": True, "use_logits": use_logits, "domains": disc_domains},
        )
        gen_f_hook = c_f.default(
            gen_f_hook,
            FeaturesForDomainLossHook,
            {"use_logits": use_logits, "domains": gen_domains},
        )
        c_hook = c_f.default(c_hook, CLossHook, {})
        disc_hook = c_f.default(
            disc_hook,
            DomainLossHook,
            {
                "d_loss_fn": disc_domain_loss_fn,
                "loss_prefix": "d_",
                "detach_features": True,
                "f_hook": disc_f_hook,
                "d_hook": disc_d_hook,
                "domains": disc_domains,
            },
        )
        gen_hook = c_f.default(
            gen_hook,
            DomainLossHook,
            {
                "d_loss_fn": gen_domain_loss_fn,
                "loss_prefix": "g_",
                "reverse_labels": True,
                "f_hook": gen_f_hook,
                "d_hook": gen_d_hook,
                "domains": gen_domains,
            },
        )
        # use gen_f_hook to get undetached features first
        disc_hook = ChainHook(
            *pre_d,
            gen_f_hook,
            disc_hook,
            *post_d,
            conditions=disc_conditions,
            alts=disc_alts
        )
        gen_hook = ChainHook(
            *pre_g, gen_hook, c_hook, *post_g, conditions=gen_conditions, alts=gen_alts
        )

        disc_hook = OptimizerHook(disc_hook, d_opts, d_weighter, d_reducer)
        gen_hook = OptimizerHook(gen_hook, g_opts, g_weighter, g_reducer)
        s_hook = SummaryHook({"d_loss": disc_hook, "g_loss": gen_hook})
        self.hook = ChainHook(disc_hook, gen_hook, s_hook)



import sys

import setuptools

sys.path.insert(0, "src")
import pytorch_adapt

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="pytorch-adapt",
    version=pytorch_adapt.__version__,
    author="Kevin Musgrave",
    description="",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KevinMusgrave/pytorch-adapt",
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.0",
    install_requires=[],
    extras_require={},
)



import torch

from ..utils import common_functions as c_f


def batch_spectral_loss(x, k):
    singular_values = torch.linalg.svdvals(x)
    return torch.sum(singular_values[:k] ** 2)


class BatchSpectralLoss(torch.nn.Module):
    # k is the number of singular values to include in the loss
    def __init__(self, k=1):
        super().__init__()
        self.k = k

    def forward(self, x):
        return batch_spectral_loss(x, self.k)

    def extra_repr(self):
        return c_f.extra_repr(self, ["k"])



class ForwardOnlyValidator:
    def run(self, adapter, **kwargs):
        if "validator" not in kwargs:
            raise KeyError(
                "An adapter validator is required when using ForwardOnlyValidator"
            )
        return adapter.run(**kwargs)



from .abs_loss import AbsLoss
from .adabn_model import AdaBNModel
from .adaptive_batch_norm import AdaptiveBatchNorm2d, PopulationBatchNorm2d
from .adaptive_feature_norm import AdaptiveFeatureNorm, L2PreservedDropout
from .batch_spectral_loss import BatchSpectralLoss, batch_spectral_loss
from .bnm_loss import BNMLoss
from .concat_softmax import ConcatSoftmax
from .confidence_weights import ConfidenceWeights
from .coral_loss import CORALLoss
from .diversity_loss import DiversityLoss
from .do_nothing_optimizer import DoNothingOptimizer
from .entropy_loss import EntropyLoss
from .entropy_weights import EntropyWeights
from .gradient_reversal import GradientReversal
from .mcc_loss import MCCLoss
from .mcd_loss import GeneralMCDLoss, MCDLoss
from .mean_dist_loss import MeanDistLoss
from .mmd_loss import MMDLoss
from .model_with_bridge import ModelWithBridge
from .multiple_models import MultipleModels
from .neighborhood_aggregation import NeighborhoodAggregation
from .normalizers import MinMaxNormalizer, SumNormalizer
from .plus_residual import PlusResidual
from .randomized_dot_product import RandomizedDotProduct
from .silhouette_score import SilhouetteScore
from .sliced_wasserstein import SlicedWasserstein
from .stochastic_linear import StochasticLinear
from .sufficient_accuracy import SufficientAccuracy
from .symnets_category_loss import SymNetsCategoryLoss, SymNetsCategoryLossListInput
from .symnets_domain_loss import SymNetsDomainLoss
from .symnets_entropy_loss import SymNetsEntropyLoss, SymNetsEntropyLossListInput
from .uniform_distribution_loss import UniformDistributionLoss
from .vat_loss import VATLoss



import torch
import torch.nn.functional as F
from pytorch_metric_learning.utils import common_functions as pml_cf

from ..utils import common_functions as c_f

# references:
# https://github.com/takerum/vat_chainer/blob/master/source/chainer_functions/loss.py
# https://github.com/takerum/vat_tf/blob/master/vat.py
# https://github.com/RuiShu/dirt-t/blob/master/codebase/models/extra_layers.py
# https://github.com/lyakaap/VAT-pytorch
# https://github.com/9310gaurav/virtual-adversarial-training


def get_normalized_noise(noise):
    original_shape = noise.shape
    noise = F.normalize(noise.view(original_shape[0], -1), dim=1)
    return noise.view(*original_shape)


class VATLoss(torch.nn.Module):
    def __init__(self, num_power_iterations=1, xi=1e-6, epsilon=8.0):
        super().__init__()
        self.num_power_iterations = num_power_iterations
        self.xi = xi
        self.epsilon = epsilon
        self.kl_div = torch.nn.KLDivLoss(reduction="batchmean")
        pml_cf.add_to_recordable_attributes(
            self, list_of_names=["num_power_iterations", "xi", "epsilon"]
        )

    def forward(self, imgs, logits, model):
        logits = logits.detach()
        model.apply(c_f.set_layers_mode("eval", c_f.batchnorm_types()))
        perturbation = self.get_perturbation(imgs, logits, model)
        new_logits = model(imgs + perturbation)

        preds = F.softmax(logits, dim=1)
        new_preds = F.log_softmax(new_logits, dim=1)
        model.apply(c_f.set_layers_mode("train", c_f.batchnorm_types()))
        return self.kl_div(new_preds, preds)

    def get_perturbation(self, imgs, original_logits, model):
        noise = torch.randn(*imgs.shape, device=original_logits.device)
        original_preds = F.softmax(original_logits, dim=1)

        for _ in range(self.num_power_iterations):
            noise.requires_grad = True
            noise = self.xi * get_normalized_noise(noise)
            noise.retain_grad()
            new_preds = F.log_softmax(model(imgs + noise), dim=1)
            dist = self.kl_div(new_preds, original_preds)
            dist.backward(retain_graph=True)
            noise = noise.grad.detach()
            model.zero_grad()

        return self.epsilon * get_normalized_noise(noise)

    def extra_repr(self):
        return c_f.extra_repr(self, ["num_power_iterations", "xi", "epsilon"])



import re

import torch

from ..utils import common_functions as c_f
from .base import BaseHook, BaseWrapperHook
from .utils import ChainHook


class BaseFeaturesHook(BaseHook):
    def __init__(
        self,
        model_name,
        in_suffixes=None,
        out_suffixes=None,
        domains=None,
        detach=False,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.model_name = model_name
        self.domains = c_f.default(domains, ["src", "target"])
        self.init_detach_mode(detach)
        self.init_suffixes(in_suffixes, out_suffixes)

    def call(self, losses, inputs):
        outputs = {}
        for domain in self.domains:
            detach = self.check_grad_mode(domain)
            func = self.mode_detached if detach else self.mode_with_grad
            in_keys = c_f.filter(self.in_keys, f"^{domain}")
            func(inputs, outputs, domain, in_keys)

        self.check_outputs_requires_grad(outputs)
        return {}, outputs

    def check_grad_mode(self, domain):
        detach = self.detach[domain]
        if not torch.is_grad_enabled():
            if not detach:
                raise ValueError(
                    f"detach[{domain}] == {detach} but grad is not enabled"
                )
        return detach

    def check_outputs_requires_grad(self, outputs):
        for k, v in outputs.items():
            if k.endswith("detached") and c_f.requires_grad(v, does=True):
                raise TypeError(f"{k} ends with 'detached' but tensor requires grad")
            if not k.endswith("detached") and c_f.requires_grad(v, does=False):
                raise TypeError(
                    f"{k} doesn't end in 'detached' but tensor doesn't require grad"
                )

    def mode_with_grad(self, inputs, outputs, domain, in_keys):
        output_keys = c_f.filter(self._out_keys(), f"^{domain}")
        output_vals = self.get_kwargs(inputs, output_keys)
        self.add_if_new(
            outputs, output_keys, output_vals, inputs, self.model_name, in_keys, domain
        )
        return output_keys, output_vals

    def mode_detached(self, inputs, outputs, domain, in_keys):
        curr_out_keys = c_f.filter(self._out_keys(), f"^{domain}")
        self.try_existing_detachable(inputs, outputs, curr_out_keys)
        remaining_out_keys = [
            k for k in curr_out_keys if k not in set().union(inputs, outputs)
        ]
        if len(remaining_out_keys) > 0:
            output_vals = self.get_kwargs(inputs, remaining_out_keys)
            with torch.no_grad():
                self.add_if_new(
                    outputs,
                    remaining_out_keys,
                    output_vals,
                    inputs,
                    self.model_name,
                    in_keys,
                    domain,
                )

    def add_if_new(
        self, outputs, full_key, output_vals, inputs, model_name, in_keys, domain
    ):
        c_f.add_if_new(outputs, full_key, output_vals, inputs, model_name, in_keys)

    def create_keys(self, domain, suffix, starting_keys=None, detach=False):
        if starting_keys is None:
            full_keys = [f"{domain}{x}" for x in suffix]
        else:
            if len(starting_keys) > 1:
                starting_keys = self.join_keys(starting_keys)
            if len(suffix) > 1:
                starting_keys = starting_keys * len(suffix)
            full_keys = [f"{k}{x}" for k, x in zip(starting_keys, suffix)]
        if detach:
            full_keys = self.add_detached_string(full_keys)
        return full_keys

    def get_kwargs(self, inputs, keys):
        return [inputs.get(k) for k in keys]

    def try_existing_detachable(self, inputs, outputs, curr_out_keys):
        for k in curr_out_keys:
            if k in inputs or k in outputs:
                continue
            curr_regex = self.detachable_regex[k]
            success = self.try_existing_detachable_in_dict(
                curr_regex, inputs, outputs, k
            )
            if not success:
                self.try_existing_detachable_in_dict(curr_regex, outputs, outputs, k)

    def try_existing_detachable_in_dict(self, regex, in_dict, outputs, new_k):
        for k, v in in_dict.items():
            if regex.search(k) and v is not None:
                outputs[new_k] = v.detach()
                return True
        return False

    def find_detachable(self, keys):
        keys = [k.replace("_", "")]

    def add_detached_string(self, keys):
        # delete existing detached string, then append to the very end
        # for example, if computing detached logits for: src_imgs_features_detached
        # 1. src_imgs_features_detached_logits --> src_imgs_features_logits
        # 2. src_imgs_features_logits --> src_imgs_features_logits_detached
        keys = [k.replace("_detached", "") for k in keys]
        return [f"{k}_detached" for k in keys]

    def join_keys(self, keys):
        return ["_AND_".join(keys)]

    def init_detach_mode(self, detach):
        if isinstance(detach, dict):
            if any(not isinstance(v, bool) for v in detach.values()):
                raise TypeError("if detach is a dict, values must be bools")
            self.detach = detach
        elif isinstance(detach, bool):
            self.detach = {k: detach for k in self.domains}
        else:
            raise TypeError("detach must be a bool or a dict of bools")

    def init_suffixes(self, in_suffixes, out_suffixes):
        self.in_suffixes = in_suffixes
        self.out_suffixes = out_suffixes
        in_keys = []
        for domain in self.domains:
            in_keys.extend(self.create_keys(domain, in_suffixes))
        self.set_in_keys(in_keys)

    def set_in_keys(self, in_keys):
        super().set_in_keys(in_keys)
        self.all_out_keys = []
        for domain in self.domains:
            curr_in_keys = c_f.filter(self.in_keys, f"^{domain}")
            curr_out_keys = self.create_keys(
                domain, self.out_suffixes, curr_in_keys, detach=self.detach[domain]
            )
            self.all_out_keys.extend(curr_out_keys)

        # strings with '_detached' optional and anywhere
        self.detachable_regex = {
            k: re.compile(
                f"^{k.replace('_detached', '').replace('_', '(_detached)?_')}$"
            )
            for k in self.all_out_keys
        }

    def _loss_keys(self):
        return []

    def _out_keys(self):
        return self.all_out_keys

    def extra_repr(self):
        return c_f.extra_repr(self, ["model_name", "domains", "detach"])


class FeaturesHook(BaseFeaturesHook):
    def __init__(
        self,
        model_name="G",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs"])
        out_suffixes = c_f.default(out_suffixes, ["_features"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class LogitsHook(BaseFeaturesHook):
    def __init__(
        self,
        model_name="C",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features"])
        out_suffixes = c_f.default(out_suffixes, ["_logits"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class FeaturesChainHook(ChainHook):
    def __init__(
        self,
        *hooks,
        **kwargs,
    ):
        for i in range(len(hooks) - 1):
            hooks[i + 1].set_in_keys(hooks[i].out_keys)
        super().__init__(*hooks, **kwargs)


class FeaturesAndLogitsHook(FeaturesChainHook):
    def __init__(
        self,
        domains=None,
        detach_features=False,
        detach_logits=False,
        other_hooks=None,
        **kwargs,
    ):
        features_hook = FeaturesHook(detach=detach_features, domains=domains)
        logits_hook = LogitsHook(detach=detach_logits, domains=domains)
        other_hooks = c_f.default(other_hooks, [])
        super().__init__(features_hook, logits_hook, *other_hooks, **kwargs)


class FeaturesWithGradAndDetachedHook(BaseWrapperHook):
    def __init__(
        self,
        model_name="G",
        in_suffixes=None,
        out_suffixes=None,
        domains=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        hooks = []
        for detach in [False, True]:
            hooks.append(
                FeaturesHook(
                    model_name=model_name,
                    in_suffixes=in_suffixes,
                    out_suffixes=out_suffixes,
                    domains=domains,
                    detach=detach,
                    **kwargs,
                )
            )
        self.hook = ChainHook(*hooks)


class CombinedFeaturesHook(BaseFeaturesHook):
    def __init__(
        self,
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(
            in_suffixes, ["_imgs_features", "_imgs_features_logits"]
        )
        out_suffixes = c_f.default(out_suffixes, ["_combined"])
        super().__init__(
            model_name="feature_combiner",
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class DLogitsHook(BaseFeaturesHook):
    def __init__(
        self,
        model_name="D",
        in_suffixes=None,
        out_suffixes=None,
        **kwargs,
    ):
        in_suffixes = c_f.default(in_suffixes, ["_imgs_features"])
        out_suffixes = c_f.default(out_suffixes, ["_dlogits"])
        super().__init__(
            model_name=model_name,
            in_suffixes=in_suffixes,
            out_suffixes=out_suffixes,
            **kwargs,
        )


class FrozenModelHook(BaseWrapperHook):
    def __init__(self, hook, model_name, **kwargs):
        super().__init__(**kwargs)
        self.hook = hook
        self.model_name = model_name

    def call(self, losses, inputs):
        model = inputs[self.model_name]
        model.eval()
        with torch.no_grad():
            losses, outputs = self.hook(losses, inputs)
        return losses, outputs
